"""
implementation of multi factor ewm model
"""
# packages
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from dataclasses import dataclass
from typing import Dict, Optional, Tuple, Literal
from enum import Enum

# qis
import qis.utils.df_ops as dfo
import qis.perfstats.returns as ret
import qis.plots.time_series as pts
import qis.models.linear.ewm as ewm
from qis.models.linear.ewm import MeanAdjType, InitType
from qis import TimePeriod


@dataclass
class LinearModel:
    """
    core class to store and process of data for different linear signals of shape:
    y =  loadings^T @ x
    """
    x: pd.DataFrame  # t, x_n factors
    y: pd.DataFrame  # t, y_m factors
    loadings: Dict[str, pd.DataFrame] = None  # estimated factor loadings

    def __post_init__(self):
        if self.loadings is not None:
            str_factors = list(self.loadings.keys())
            assert str_factors == self.x.columns.to_list()

    def print(self):
        print(f"x:\n{self.x}")
        print(f"y:\n{self.y}")
        for factor, loading in self.loadings.items():
            print(f"{factor}:\n{loading}")

    def get_factor_loadings(self, factor: str) -> pd.DataFrame:
        return self.loadings[factor]

    def compute_agg_factor_exposures(self,
                                     exposures: pd.DataFrame
                                     ) -> pd.DataFrame:
        factor_exposures = {}
        for factor, loading in self.loadings.items():
            factor_exposures[factor] = loading.multiply(exposures).sum(axis=1)
        factor_exposures = pd.DataFrame.from_dict(factor_exposures)
        return factor_exposures

    def get_asset_factor_betas(self, asset: str = None) -> pd.DataFrame:
        """
        return df of asset exposures to factors
        """
        if asset is None:
            asset = self.y.columns[0]
        exps = {}
        for factor, factor_exp in self.loadings.items():
            exps[factor] = factor_exp[asset]
        exps = pd.DataFrame.from_dict(exps)
        return exps

    def get_asset_factor_attribution(self, asset: str = None, add_total: bool = True) -> pd.DataFrame:
        """
        return df of asset exposures to factors
        """
        factor_betas = self.get_asset_factor_betas(asset=asset)
        exposures = self.x
        attribution = exposures.multiply(factor_betas.shift(1))
        if add_total:
            total = attribution.sum(1).rename('Total')
            attribution = pd.concat([total, attribution], axis=1)
        return attribution

    def get_factor_alpha(self, lag: Literal[0, 1] = 1) -> Tuple[pd.DataFrame, pd.DataFrame]:
        """
        factor_alpha = y - sum(factor_beta_{t-shift}*x)
        lag = 1 for step-1 forecast
        lag = 0 for insample forecast
        """
        explained_returns = pd.DataFrame(0.0, index=self.y.index, columns=self.y.columns)
        for factor in self.x.columns:
            factor_betas = self.loadings[factor]
            explained_return = (factor_betas.shift(lag)).multiply(self.x[factor].to_numpy(), axis=0)
            explained_returns = explained_returns.add(explained_return)
        factor_alpha = self.y.subtract(explained_returns)
        return factor_alpha, explained_returns

    def get_model_ewm_r2(self, span: int = 52, lag: Literal[0, 1] = 0) -> pd.DataFrame:
        """
        ss_res = ewm (y - sum(factor_beta_{t-1}*x)) ^ 2
        """
        residuals, explained_returns = self.get_factor_alpha(lag=lag)
        ewm_residuals_2 = ewm.compute_ewm(data=np.square(residuals), span=span)
        y_demean = self.y #- ewm.compute_ewm(data=self.y, span=span)
        ewm_variance = ewm.compute_ewm(data=np.square(y_demean), span=span)
        r_2 = 1.0 - ewm_residuals_2.divide(ewm_variance)
        return r_2

    def get_model_residuals_corrs(self, span: int = 52) -> Tuple[pd.DataFrame, pd.Series]:
        residuals, explained_returns = self.get_factor_alpha(lag=0)
        corr = ewm.compute_ewm_covar(residuals.to_numpy(), span=span, is_corr=True)
        avg_corr = pd.Series(0.5*np.nanmean(corr - np.eye(corr.shape[0]), axis=1), index=self.y.columns)
        corr_pd = pd.DataFrame(corr, index=self.y.columns, columns=self.y.columns)
        return corr_pd, avg_corr

    def plot_factor_loadings(self,
                             factor: str,
                             var_format: str = '{:,.2f}',
                             time_period: TimePeriod = None,
                             ax: plt.Subplot = None,
                             **kwargs):
        df = self.loadings[factor]
        if time_period is not None:
            df = time_period.locate(df)
        pts.plot_time_series(df=df,
                             var_format=var_format,
                             ax=ax, **kwargs)


class EwmLinearModel(LinearModel):
    """
    implementation for ewma model
    """
    def fit(self,
            span: Optional[int] = 31,
            ewm_lambda: float = 0.94,
            is_x_correlated: bool = True,
            mean_adj_type: MeanAdjType = MeanAdjType.NONE,
            init_type: InitType = InitType.MEAN
            ) -> None:
        """
        estimate time series ewm betas
        factor_returns is factors returns data: T*M
        asset_returns is asset returns data: T*N
        return type is {factor_returns.columns: factor of asset_returns T*N}

        is_independent specifies the asset_returns covar structure
        is_x_correlated:
        True = x covariance is orthogonal with ewm_covar being diagonal matrix
        False = x covariance is ewm covariance with ewm_covar being full matrix

        ewm betas = (ewm_covar[x, x]^{-1}) @ (ewm_covar[x, y])
        """
        x = self.x
        y = self.y
        if span is not None:
            ewm_lambda = 1.0 - 2.0 / (span + 1.0)
        if mean_adj_type != MeanAdjType.NONE:
            x = ewm.compute_rolling_mean_adj(data=x,
                                             mean_adj_type=mean_adj_type,
                                             ewm_lambda=ewm_lambda,
                                             init_type=init_type)

            y = ewm.compute_rolling_mean_adj(data=y,
                                             mean_adj_type=mean_adj_type,
                                             ewm_lambda=ewm_lambda,
                                             init_type=init_type)

        # compute list of betas using ewm numba recursion for cross product of x y and covariance of x
        # output is tensor of betas per date = [t, factors, assets]
        betas_ts = ewm.compute_ewm_xy_beta_tensor(x=x.to_numpy(),
                                                  y=y.to_numpy(),
                                                  ewm_lambda=ewm_lambda,
                                                  is_x_correlated=is_x_correlated)
        # factor_loadings = {factor_id: pd.DataFrame(factor loadings)}
        loadings = dfo.np_txy_tensor_to_pd_dict(np_tensor_txy=betas_ts,
                                                dateindex=x.index,
                                                factor_names=x.columns.to_list(),
                                                asset_names=y.columns.to_list())
        self.x = x
        self.y = y
        self.loadings = loadings


def compute_portfolio_benchmark_betas(instrument_prices: pd.DataFrame,
                                      exposures: pd.DataFrame,
                                      benchmark_prices: pd.DataFrame,
                                      time_period: TimePeriod = None,
                                      beta_freq: str = None,
                                      factor_beta_span: int = 63  # quarter
                                      ) -> pd.DataFrame:
    """
    compute benchmark betas of instruments
    portfolio_beta_i = sum(instrument_beta_i*exposure)
    """
    benchmark_prices = benchmark_prices.reindex(index=instrument_prices.index, method='ffill')
    ewm_linear_model = EwmLinearModel(x=ret.to_returns(prices=benchmark_prices, freq=beta_freq, is_log_returns=True),
                                      y=ret.to_returns(prices=instrument_prices, freq=beta_freq, is_log_returns=True))
    ewm_linear_model.fit(span=factor_beta_span, is_x_correlated=True)
    exposures = exposures.reindex(index=instrument_prices.index, method='ffill')
    benchmark_betas = ewm_linear_model.compute_agg_factor_exposures(exposures=exposures)
    benchmark_betas = benchmark_betas.replace({0.0: np.nan}).ffill()  # fillholidays
    if time_period is not None:
        benchmark_betas = time_period.locate(benchmark_betas)
    return benchmark_betas


def compute_portfolio_benchmark_beta_alpha_attribution(instrument_prices: pd.DataFrame,
                                                       exposures: pd.DataFrame,
                                                       benchmark_prices: pd.DataFrame,
                                                       portfolio_nav: pd.Series,
                                                       time_period: TimePeriod = None,
                                                       beta_freq: str = None,
                                                       factor_beta_span: int = 63,  # quarter
                                                       residual_name: str = 'Alpha'
                                                       ) -> pd.DataFrame:
    """
    attribution:=alpha_{t} = portfolio_return_{t} - benchmark_return_{t}*beta_{t-1}
    using compounded returns
    portfolio_nav is the gross/net portfolio nav
    """
    portfolio_benchmark_betas = compute_portfolio_benchmark_betas(instrument_prices=instrument_prices,
                                                                  exposures=exposures,
                                                                  benchmark_prices=benchmark_prices,
                                                                  time_period=None,
                                                                  beta_freq=beta_freq,
                                                                  factor_beta_span=factor_beta_span)
    joint_attrib = compute_benchmarks_beta_attribution(portfolio_nav=portfolio_nav,
                                                       benchmark_prices=benchmark_prices,
                                                       portfolio_benchmark_betas=portfolio_benchmark_betas,
                                                       residual_name=residual_name,
                                                       time_period=time_period)
    return joint_attrib


def compute_benchmarks_beta_attribution(portfolio_nav: pd.Series,
                                        benchmark_prices: pd.DataFrame,
                                        portfolio_benchmark_betas: pd.DataFrame,
                                        residual_name: str = 'Alpha',
                                        time_period: TimePeriod = None
                                        ) -> pd.DataFrame:
    benchmark_prices = benchmark_prices.reindex(index=portfolio_benchmark_betas.index, method='ffill')
    portfolio_nav = portfolio_nav.reindex(index=portfolio_benchmark_betas.index, method='ffill')
    x = ret.to_returns(prices=benchmark_prices, freq=None)
    x_attribution = (portfolio_benchmark_betas.shift(1)).multiply(x)
    total_attrib = x_attribution.sum(axis=1)
    total = portfolio_nav.pct_change()
    residual = np.subtract(total, total_attrib)
    joint_attrib = pd.concat([x_attribution, residual.rename(residual_name)], axis=1)
    if time_period is not None:
        joint_attrib = time_period.locate(joint_attrib)
    return joint_attrib


class UnitTests(Enum):
    MODEL = 1
    ATTRIBUTION = 2


def run_unit_test(unit_test: UnitTests):

    from qis.test_data import load_etf_data
    prices = load_etf_data().dropna()

    if unit_test == UnitTests.MODEL:
        returns = np.log(prices.divide(prices.shift(1)))

        # factors
        factors = ['SPY', 'TLT', 'GLD']
        # factors = ['SPY']
        factor_returns = returns[factors]

        # assets
        is_check = False
        if is_check:
            asset_returns = returns[factors]
            asset_returns.columns = [f"{x.split('_')[0]}_asset" for x in factors]
        else:
            assets = ['QQQ', 'HYG']
            asset_returns = returns[assets]
        ewm_linear_model = EwmLinearModel(x=factor_returns, y=asset_returns)
        ewm_linear_model.fit(ewm_lambda=0.94, is_x_correlated=True)

        ewm_linear_model.print()
        ewm_linear_model.plot_factor_loadings(factor='SPY')

        factor_alpha, explained_returns = ewm_linear_model.get_factor_alpha()
        pts.plot_time_series(df=factor_alpha.cumsum(0), title='Cumulative alpha')
        pts.plot_time_series(df=explained_returns.cumsum(0), title='Cumulative explained return')

    elif unit_test == UnitTests.ATTRIBUTION:
        benchmark_prices = prices[['SPY', 'TLT']]
        instrument_prices = prices[['QQQ', 'HYG', 'GLD']]
        exposures = pd.DataFrame(1.0/3.0, index=instrument_prices.index, columns=instrument_prices.columns)
        portfolio_nav = ret.returns_to_nav(returns=(exposures.shift(1)).multiply(instrument_prices.pct_change()).sum(axis=1))
        print(portfolio_nav)

        attribution = compute_portfolio_benchmark_beta_alpha_attribution(instrument_prices=instrument_prices,
                                                                         exposures=exposures,
                                                                         benchmark_prices=benchmark_prices,
                                                                         portfolio_nav=portfolio_nav,
                                                                         time_period=None,
                                                                         beta_freq='W-WED',
                                                                         factor_beta_span=52,  # quarter
                                                                         residual_name='Alpha')
        pts.plot_time_series(df=attribution.cumsum(0))

    plt.show()


if __name__ == '__main__':

    unit_test = UnitTests.MODEL

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)




# packages
import numpy as np
import pandas as pd
from typing import Tuple, Union, List

# qis
import qis.utils as qu
import qis.perfstats.returns as ret
import qis.models.linear.ewm as ewm


def simulate_vol_target_strats(prices: Union[pd.DataFrame, pd.Series],
                               vol_span: int = 21,
                               vol_target: float = 0.15,
                               constant_trade_level: bool = False,
                               vol_af: float = 260
                               ) -> Tuple[pd.DataFrame, pd.DataFrame]:
    """
    simulate weights and returns on vol target
    """
    log_returns = ret.to_returns(prices=prices, is_log_returns=True)
    returns = ret.to_returns(prices=prices, is_log_returns=False)
    ewm_vol = ewm.compute_ewm_vol(data=log_returns,
                                  span=vol_span,
                                  mean_adj_type=ewm.MeanAdjType.NONE,
                                  af=vol_af)
    # vol target weights
    weights_100 = qu.to_finite_reciprocal(data=ewm_vol, fill_value=0.0, is_gt_zero=True)
    nav_weights = weights_100.multiply(vol_target)
    vt_returns = returns.multiply(nav_weights.shift(1))
    vt_navs = ret.returns_to_nav(returns=vt_returns, constant_trade_level=constant_trade_level)
    return nav_weights, vt_navs


def simulate_vol_target_strats_range(prices: Union[pd.DataFrame, pd.Series],
                                     vol_spans: List[int] = (21, 31),
                                     vol_target: float = 0.15,
                                     constant_trade_level: bool = False,
                                     vol_af: float = 260,
                                     add_asset: bool = True
                                     ) -> Tuple[pd.DataFrame, pd.DataFrame]:
    vt_nav_weights, vt_navs = [], []
    for vol_span in vol_spans:
        vt_nav_weights_, vt_navs_ = simulate_vol_target_strats(prices=prices, vol_span=vol_span, vol_target=vol_target,
                                                               constant_trade_level=constant_trade_level, vol_af=vol_af)
        if isinstance(prices, pd.Series):
            name = f"{prices.name} vol_span={vol_span}"
            vt_nav_weights_.name, vt_navs_.name = name, name
        else:
            names = [f"{x} vol_span={vol_span}" for x in prices.columns]
            vt_nav_weights_.columns, vt_navs_.columns = names, names
        vt_nav_weights.append(vt_nav_weights_)
        vt_navs.append(vt_navs_)
    vt_nav_weights, vt_navs = pd.concat(vt_nav_weights, axis=1), pd.concat(vt_navs, axis=1)
    if add_asset:
        vt_navs = pd.concat([prices, vt_navs], axis=1)
    return vt_nav_weights, vt_navs


def simulate_trend_starts(prices: Union[pd.DataFrame, pd.Series],
                          vol_span: int = 31,
                          tf_span: int = 63,
                          vol_target: float = 0.15,
                          constant_trade_level: bool = False,
                          vol_af: float = 260
                          ) -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]:
    """
    simulate weights and returns on tf strats
    """
    log_returns = ret.to_returns(prices=prices, is_log_returns=True)
    returns = ret.to_returns(prices=prices)

    ewm_vol = ewm.compute_ewm_vol(data=log_returns, span=vol_span, mean_adj_type=ewm.MeanAdjType.NONE,
                                  annualize=False)

    # vol target weights
    weights_100 = qu.to_finite_reciprocal(data=ewm_vol, fill_value=0.0, is_gt_zero=True)
    vt_return_100 = returns.multiply(weights_100.shift(1))
    # signal is unit var
    signals = ewm.compute_ewm(data=vt_return_100, span=tf_span, is_unit_vol_scaling=True)
    # normalized to target vol
    weights = signals.multiply(weights_100).multiply(vol_target/np.sqrt(vol_af))
    vt_returns = returns.multiply(weights.shift(1))
    vt_navs = ret.returns_to_nav(returns=vt_returns, constant_trade_level=constant_trade_level)
    return weights, vt_navs, signals


def simulate_trend_starts_range(prices: Union[pd.DataFrame, pd.Series],
                                vol_span: int = 31,
                                tf_spans: List[int] = (21, 63),
                                vol_target: float = 0.15,
                                constant_trade_level: bool = False,
                                vol_af: float = 260,
                                add_asset: bool = True
                                ) -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]:
    tf_nav_weights, tf_navs, signals = [], [], []
    for tf_span in tf_spans:
        tf_nav_weights_, tf_navs_, signals_ = simulate_trend_starts(prices=prices, tf_span=tf_span, vol_span=vol_span,
                                                                    vol_target=vol_target, vol_af=vol_af,
                                                                    constant_trade_level=constant_trade_level)
        if isinstance(prices, pd.Series):
            name = f"{prices.name} tf_span={tf_span}"
            tf_nav_weights_.name, tf_navs_.name, signals_.name = name, name, name
        else:
            names = [f"{x} tf_span={tf_span}" for x in prices.columns]
            tf_nav_weights_.columns, tf_navs_.columns, signals_.columns = names, names, names
        tf_nav_weights.append(tf_nav_weights_)
        tf_navs.append(tf_navs_)
        signals.append(signals_)
    tf_nav_weights, tf_navs, signals = pd.concat(tf_nav_weights, axis=1), pd.concat(tf_navs, axis=1), pd.concat(signals, axis=1)
    if add_asset:
        tf_navs = pd.concat([prices, tf_navs], axis=1)
    return tf_nav_weights, tf_navs, signals



"""
implementation of exponentially weighted mean (ewm) filter using numpy and lambda
"""
# packages
import warnings
import numpy as np
import pandas as pd
from numba import njit
from typing import Union, Tuple, Optional
from enum import Enum

# qis
import qis.utils.dates as da
import qis.utils.np_ops as npo


class NanBackfill(Enum):
    """
    when you sung ewm recursion we need to treat nans
    base case: when we assume conitnuous time series with occasional gaps we use FFILL method
    case where outputs needs a flag that given data are nans we need to use nan fill
    """
    FFILL = 1   # use last nonnan value
    DEFLATED_FFILL = 2  #  use last nonnan value * lambda
    ZERO_FILL = 3  # use zero value: nans must be filled by zero otherwise the recursion cannot start
    NAN_FILL = 4  # use nan value: for recursion we use ZERO_FILL then we substitute zeros with nans:
    # it corresponds to DEFLATED_FFILL for subsequent non nans


class InitType(Enum):
    ZERO = 1
    X0 = 2
    MEAN = 3
    VAR = 4


class MeanAdjType(Enum):
    NONE = 1
    INSAMPLE = 2
    EXPANDING = 3
    EWMA = 4


class CrossXyType(Enum):
    COVAR = 1
    BETA = 2
    CORR = 3


def set_init_dim1(data: Union[pd.DataFrame, pd.Series, np.ndarray],
                  init_type: InitType = InitType.X0
                  ) -> Union[float, np.ndarray]:

    x = npo.to_finite_np(data=data, fill_value=np.nan)

    if init_type == InitType.ZERO:
        init_value = np.zeros_like(x[0])
    elif init_type == InitType.X0:
        init_value = np.where(np.isnan(x[0])==False, x[0], 0.0)
    elif init_type == InitType.MEAN:
        init_value = np.nanmean(x, axis=0)
        init_value = np.where(np.isnan(init_value) == False, init_value, 0.0)
    elif init_type == InitType.VAR:
        init_value = np.nanvar(x, axis=0)
    else:
        raise TypeError(f"in set_initial_condition: unsuported init_type")

    return init_value


def set_init_dim2(data: Union[pd.DataFrame, pd.Series, np.ndarray],
                  init_type: InitType = InitType.X0
                  ) -> np.ndarray:

    x = npo.to_finite_np(data=data, fill_value=np.nan)
    cross = np.transpose(x) @ x
    if init_type == InitType.ZERO:
        init_value = np.zeros((cross.shape[0], cross.shape[1]))
    elif init_type == InitType.X0:
        init_value = np.zeros((cross.shape[0], cross.shape[1]))
    else:
        raise TypeError(f"in set_initial_condition_dim2: unsuported init_type")

    return init_value


@njit
def ewm_recursion(a: np.ndarray,
                  init_value: Union[float, np.ndarray],
                  span: Union[int, np.ndarray] = None,
                  ewm_lambda: Union[float, np.ndarray] = 0.94,
                  is_start_from_first_nonan: bool = True,
                  is_unit_vol_scaling: bool = False,
                  nan_backfill: NanBackfill = NanBackfill.FFILL
                  ) -> np.ndarray:

    """
    compute ewm using recursion:
    ewm[t] = (1-lambda) * x[t] + lambda*ewm[t-1]

    assumption is that non np.nan value is returned from the function

    data: numpy with dimension = t*n
    ewm_lambda: float or ndarray of dimension n
    init_value: initial value of dimension n
    start_from_first_nonan: start filling nans only from the first non-nan in underlying data: recomended because
                            it avoids backfilling of init_value
    is_unit_vol_scaling: outputs are scaled to have var(ewm)=1 (for gaussian data with zero corrs)
    """
    if span is not None:
        ewm_lambda = 1.0 - 2.0 / (span + 1.0)
    ewm_lambda_1 = 1.0 - ewm_lambda

    is_1d = (a.ndim == 1)  # or a.shape[1] == 1)

    # initialize all
    ewm = np.full_like(a, fill_value=np.nan, dtype=np.double)

    if is_start_from_first_nonan:
        if is_1d:  # cannot use np.where
            last_ewm = init_value if np.isfinite(a[0]) else np.nan
        else:
            last_ewm = np.where(np.isfinite(a[0]), init_value, np.nan)
    else:
        last_ewm = init_value
    ewm[0] = last_ewm

    # recurse from 1
    for t in np.arange(1, a.shape[0]):
        a_t = a[t]

        if is_start_from_first_nonan:
            # detect starting nonnans for when last ewma was np.nan and a_t is finite
            if is_1d:  # cannot use np.where
                if np.isfinite(last_ewm)==False and np.isfinite(a_t)==True:  # trick: if last_ewm is nan
                    last_ewm = init_value
            else:
                new_nonnans = np.logical_and(np.isfinite(last_ewm)==False, np.isfinite(a_t)==True)
                if np.any(new_nonnans):
                    last_ewm = np.where(new_nonnans, init_value, last_ewm)

        # do the step
        current_ewm = ewm_lambda * last_ewm + ewm_lambda_1 * a_t

        # fill nan-values
        if is_1d:   # np.where cannot be used
            if not np.isfinite(current_ewm):
                if nan_backfill == NanBackfill.FFILL:
                    current_ewm = last_ewm
                elif nan_backfill == NanBackfill.DEFLATED_FFILL:
                    current_ewm = ewm_lambda*last_ewm
                else:  # use zero fill
                    current_ewm = 0.0
        else:
            if nan_backfill == NanBackfill.FFILL:
                fill_value = last_ewm
            elif nan_backfill == NanBackfill.DEFLATED_FFILL:
                fill_value = ewm_lambda*last_ewm
            else:  # use zero fill
                fill_value = np.zeros_like(last_ewm)

            current_ewm = np.where(np.isfinite(current_ewm), current_ewm, fill_value)

        ewm[t] = last_ewm = current_ewm

    if is_unit_vol_scaling:
        vol_ratio = np.sqrt(1 - ewm_lambda) / (1 - np.sqrt(ewm_lambda))
        ewm = vol_ratio * ewm

    return ewm


def compute_ewm_long_short_filter(data: pd.DataFrame,
                                  long_span: Union[int, np.ndarray] = 63,
                                  short_span: Optional[Union[int, np.ndarray]] = 5,
                                  warmup_period: Optional[Union[int, np.ndarray]] = 21
                                  ) -> pd.DataFrame:
    """
    signal smoother
    """
    long_lambda = 1.0 - 2.0 / (long_span+1.0)

    data_np = data.to_numpy()
    init_value = np.zeros(data_np.shape[1])
    if short_span is not None:  # use short + long filter
        short_lambda = 1.0 - 2.0 / (short_span + 1.0)
        weight_long = 1.0 / (np.sqrt(1.0 - long_lambda) * np.sqrt(1.0 / (1.0 - long_lambda)
                                                                  + 1.0 / (1.0 - short_lambda)
                                                                  - 2.0 / (1.0 - np.sqrt(long_lambda * short_lambda))))
        weight_short = 1.0 / (np.sqrt(1.0 - short_lambda) * np.sqrt(1.0 / (1.0 - long_lambda)
                                                                    + 1.0 / (1.0 - short_lambda)
                                                                    - 2.0 / (1.0 - np.sqrt(long_lambda * short_lambda))))
        load_long = np.sqrt(1.0 - long_lambda) / (1.0 - np.sqrt(long_lambda))
        long_signal = weight_long * ewm_recursion(a=load_long * data_np,
                                                  ewm_lambda=np.sqrt(long_lambda),
                                                  init_value=init_value)
        load_short = np.sqrt(1.0 - short_lambda) / (1.0 - np.sqrt(short_lambda))
        short_signal = weight_short * ewm_recursion(a=load_short * data_np,
                                                    ewm_lambda=np.sqrt(short_lambda),
                                                    init_value=init_value)
        filter = long_signal - short_signal

    else:
        load_long = np.sqrt(1.0 - long_lambda) / (1.0 - np.sqrt(long_lambda))
        filter = ewm_recursion(a=load_long * data_np,
                               ewm_lambda=np.sqrt(long_lambda),
                               init_value=init_value)

    if warmup_period is not None:   # set to nan first nonnan in warmup_period
        if isinstance(warmup_period, int):
            warmup_period = np.full(data_np.shape[1], warmup_period)
        elif isinstance(warmup_period, np.ndarray):
            pass
        else:
            raise ValueError(f"type={type(warmup_period)}")
        if np.any(warmup_period > 0):
            for idx, period in enumerate(warmup_period):
                nan_indicators = np.argwhere(np.isfinite(data_np[:, idx]))
                if nan_indicators.size > period:
                    filter[:nan_indicators[period][0], idx] = np.nan
                else:
                    filter[:, idx] = np.nan

    filter = pd.DataFrame(data=filter, index=data.index, columns=data.columns)
    return filter


@njit
def compute_ewm_covar(a: np.ndarray,
                      span: Union[int, np.ndarray] = None,
                      ewm_lambda: float = 0.94,
                      covar0: np.ndarray = None,
                      is_corr: bool = False,
                      nan_backfill: NanBackfill = NanBackfill.FFILL
                      ) -> np.ndarray:
    """
    compute ewm covariance matrix
    """

    if span is not None:
        ewm_lambda = 1.0 - 2.0 / (span + 1.0)
    ewm_lambda_1 = 1.0 - ewm_lambda

    if a.ndim == 1:  # ndarry
        n = a.shape[0]
    else:
        n = a.shape[1]  # array of ndarray

    if covar0 is None:
        covar = np.zeros((n, n))
    else:
        covar = np.where(np.isfinite(covar0), covar0, 0.0)

    last_covar = covar
    if a.ndim == 1:  # ndarry array
        r_ij = np.outer(a, a)
        covar = ewm_lambda_1 * r_ij + ewm_lambda * last_covar
        if nan_backfill == NanBackfill.FFILL:
            fill_value = last_covar
        elif nan_backfill == NanBackfill.DEFLATED_FFILL:
            fill_value = ewm_lambda * last_covar
        else:  # use zero fill
            fill_value = np.zeros_like(last_covar)

        covar = last_covar = np.where(np.isfinite(covar), covar, fill_value)

    else:  # loop over rows
        t = a.shape[0]
        for idx in range(0, t):  # row in x:
            row = a[idx]
            r_ij = np.outer(row, row)
            covar = ewm_lambda_1 * r_ij + ewm_lambda * last_covar

            if nan_backfill == NanBackfill.FFILL:
                fill_value = last_covar
            elif nan_backfill == NanBackfill.DEFLATED_FFILL:
                fill_value = ewm_lambda*last_covar
            else:  # use zero fill
                fill_value = np.zeros_like(last_covar)

            last_covar = np.where(np.isfinite(covar), covar, fill_value)

            if is_corr:
                if np.nansum(np.diag(last_covar)) > 1e-10:
                    inv_vol = np.reciprocal(np.sqrt(np.diag(last_covar)))
                    norm = np.outer(inv_vol, inv_vol)
                else:
                    norm = np.identity(n)
                last_covar_ = norm * last_covar
            else:
                last_covar_ = last_covar

            covar = last_covar_

    return covar


@njit
def compute_ewm_covar_tensor(a: np.ndarray,
                             span: Union[int, np.ndarray] = None,
                             ewm_lambda: float = 0.94,
                             covar0: np.ndarray = None,
                             is_corr: bool = False,
                             nan_backfill: NanBackfill = NanBackfill.FFILL
                             ) -> np.ndarray:
    """
    compute ewm covariance matrix time series as 3-d tensor [t, x, x]
    """
    if span is not None:
        ewm_lambda = 1.0 - 2.0 / (span + 1.0)
    ewm_lambda_1 = 1.0 - ewm_lambda

    if not a.ndim == 2:
        raise ValueError(f"only 2-d arrays are supported")

    t = a.shape[0]
    n = a.shape[1]  # array of ndarray
    zero_covar = np.zeros((n, n))

    if covar0 is None:
        covar = zero_covar
    else:
        covar = np.where(np.isfinite(covar0), covar0, zero_covar)

    output_covar = np.empty((t, n, n))
    last_covar = covar
    # loop over rows
    for idx in range(0, t):  # row in x:
        row = a[idx]
        r_ij = np.outer(row, row)
        covar = ewm_lambda_1 * r_ij + ewm_lambda * last_covar
        if nan_backfill == NanBackfill.FFILL:
            last_covar = np.where(np.isfinite(covar), covar, last_covar)
        elif nan_backfill == NanBackfill.DEFLATED_FFILL:
            last_covar = np.where(np.isfinite(covar), covar, ewm_lambda * last_covar)
        else:  # use zero fill
            last_covar = np.where(np.isfinite(covar), covar, zero_covar)

        if is_corr:
            if np.nansum(np.diag(last_covar)) > 1e-10:
                inv_vol = np.reciprocal(np.sqrt(np.diag(last_covar)))
                norm = np.outer(inv_vol, inv_vol)
            else:
                norm = np.identity(n)
            last_covar_ = norm * last_covar
        else:
            last_covar_ = last_covar

        if nan_backfill == NanBackfill.NAN_FILL:  # fill zeros with nans
            last_covar_ = np.where(np.equal(last_covar_, zero_covar), np.nan, last_covar_)

        output_covar[idx] = last_covar_

    return output_covar


@njit
def compute_ewm_covar_tensor_vol_norm_returns(a: np.ndarray,
                                              span: Union[int, np.ndarray] = None,
                                              ewm_lambda: float = 0.94,
                                              covar0: np.ndarray = None,
                                              is_corr: bool = False,
                                              nan_backfill: NanBackfill = NanBackfill.FFILL
                                              ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    compute ewm covariance matrix time series as 3-d tensor [t, x, x]
    returns vector a is normalised by vol
    compute normalised covar, covar for natural vars and vols
    """
    if span is not None:
        ewm_lambda = 1.0 - 2.0 / (span + 1.0)
    ewm_lambda_1 = 1.0 - ewm_lambda

    if not a.ndim == 2:
        raise ValueError(f"only 2-d arrays are supported")

    t = a.shape[0]
    n = a.shape[1]  # array of ndarray
    zero_covar = np.zeros((n, n))

    if covar0 is None:
        covar = zero_covar
    else:
        covar = np.where(np.isfinite(covar0), covar0, zero_covar)

    output_covar_norm = np.empty((t, n, n))
    output_covar = np.empty((t, n, n))
    last_covar = covar

    # compute vols
    a_var = np.square(a)
    ewm_vol = np.sqrt(ewm_recursion(a=a_var, ewm_lambda=ewm_lambda, init_value=np.nanmean(a_var, axis=0), nan_backfill=nan_backfill))
    a_norm = np.divide(a, ewm_vol, where=np.greater(ewm_vol, 0.0))

    # loop over rows
    for idx in range(0, t):  # row in x:
        row = a_norm[idx]
        r_ij = np.outer(row, row)
        covar = ewm_lambda_1 * r_ij + ewm_lambda * last_covar
        if nan_backfill == NanBackfill.FFILL:
            last_covar = np.where(np.isfinite(covar), covar, last_covar)
        elif nan_backfill == NanBackfill.DEFLATED_FFILL:
            last_covar = np.where(np.isfinite(covar), covar, ewm_lambda * last_covar)
        else:  # use zero fill
            last_covar = np.where(np.isfinite(covar), covar, zero_covar)

        if is_corr:
            if np.nansum(np.diag(last_covar)) > 1e-10:
                inv_vol = np.reciprocal(np.sqrt(np.diag(last_covar)))
                norm = np.outer(inv_vol, inv_vol)
            else:
                norm = np.identity(n)
            last_covar_ = norm * last_covar
        else:
            last_covar_ = last_covar

        if nan_backfill == NanBackfill.NAN_FILL:  # fill zeros with nans
            last_covar_ = np.where(np.equal(last_covar_, zero_covar), np.nan, last_covar_)

        ewm_vol_ = ewm_vol[idx]
        output_covar[idx] = last_covar_ * np.outer(ewm_vol_, ewm_vol_)
        output_covar_norm[idx] = last_covar_

    return output_covar_norm, output_covar, ewm_vol


# @njit
def compute_ewm_xy_beta_tensor(x: np.ndarray,  # factor returns
                               y: np.ndarray,  # asset returns
                               span: Union[int, np.ndarray] = None,
                               ewm_lambda: float = 0.94,
                               warm_up_period: int = 20,  # to avoid excessive betas at start,
                               is_x_correlated: bool = True,  # computation of [x,x]
                               nan_backfill: NanBackfill = NanBackfill.FFILL
                               ) -> np.ndarray:
    """
    compute ewm cross matrices with x*y using outer product = dim[x] * dim[y]
    the dimension of tensor is [t, x, y]
    njit
    """
    if not x.ndim in [1, 2] or not y.ndim in [1, 2]:
        raise TypeError("Expected 1- or 2-dimensional NumPy array for x and y")
    if x.shape[0] != y.shape[0]:
        raise TypeError("first time series dimension of x and y must be equal")

    if x.ndim == 1:  # numba is sensetive to how dimensions and initial value
        nx = 1
        is_x_correlated = False  # for 1-d factor no need to compute outer product
    else:
        nx = x.shape[1]
    inv_t1 = np.diag(np.ones(nx))
    last_covar_xx = np.zeros((nx, nx))
    if y.ndim == 1:
        ny = 1
    else:
        ny = y.shape[1]
    last_cross_xy = np.zeros((nx, ny))
    beta_nan = np.full((nx, ny), np.nan)

    nt = x.shape[0]
    betas_ts = np.full((nt, nx, ny), np.nan)

    if span is not None:
        ewm_lambda = 1.0 - 2.0 / (span + 1.0)
    ewm_lambda_1 = 1.0 - ewm_lambda
    for t in range(nt):  # over time index
        row_x = x[t]  # time series row
        row_y = y[t]
        cross_xy = ewm_lambda_1 * np.outer(row_x, row_y) + ewm_lambda * last_cross_xy

        if nan_backfill == NanBackfill.FFILL:
            fill_value = last_cross_xy
        elif nan_backfill == NanBackfill.DEFLATED_FFILL:
            fill_value = ewm_lambda * last_cross_xy
        else:  # use zero fill
            fill_value = np.zeros_like(last_cross_xy)

        cross_xy = np.where(np.isfinite(cross_xy), cross_xy, fill_value)
        last_cross_xy = cross_xy

        # covar matrix
        covar_xx = ewm_lambda_1 * np.outer(row_x, row_x) + ewm_lambda * last_covar_xx
        if nan_backfill:
            covar_xx = np.where(np.isfinite(covar_xx), covar_xx, ewm_lambda * last_covar_xx)
        else:
            covar_xx = np.where(np.isfinite(covar_xx), covar_xx, last_covar_xx)
        last_covar_xx = covar_xx

        if t > warm_up_period:
            # if np.trace(covar_xx) > 1e-8:
            if np.min(np.diag(covar_xx)) > 1e-8:
                if is_x_correlated:  # use inversion
                    try:
                        inv_t = np.linalg.inv(covar_xx)
                    except:  # "Singular matrix": #LinAlgError("Singular matrix")
                        inv_t = np.diag(np.reciprocal(np.diag(covar_xx)))
                    inv_t = np.ascontiguousarray(inv_t)  # to remove numpy warning
                else:
                    # reciprocal of diagonal elements
                    inv_t = np.diag(np.reciprocal(np.diag(covar_xx)))
            else:
                inv_t = inv_t1

            inv_t = np.where(np.isfinite(inv_t), inv_t, inv_t1)
            betas_t = inv_t @ cross_xy
        else:
            betas_t = beta_nan

        betas_ts[t] = betas_t

    return betas_ts


def compute_one_factor_ewm_betas(x: pd.Series,
                                 y: pd.DataFrame,
                                 span: Union[int, np.ndarray] = None,
                                 ewm_lambda: float = 0.94,
                                 nan_backfill: NanBackfill = NanBackfill.FFILL
                                 ) -> pd.DataFrame:
    """
    ewm betas of y wrt factor 1-d x
    """
    if not x.index.equals(y.index):
        raise ValueError("x.index={x.index} is not equal to y.index={y.index}")

    x_np = npo.to_finite_np(data=x, fill_value=np.nan)
    y_np = npo.to_finite_np(data=y, fill_value=np.nan)

    betas_ts = compute_ewm_xy_beta_tensor(x=x_np, y=y_np, span=span,
                                          ewm_lambda=ewm_lambda,
                                          nan_backfill=nan_backfill)
    # the x factor dimension is 1, we get slice [t, y] using [:, 0, :]
    one_factor_ewm_betas = pd.DataFrame(data=betas_ts[:, 0, :], index=y.index, columns=y.columns)
    return one_factor_ewm_betas


def compute_ewm(data: Union[pd.DataFrame, pd.Series, np.ndarray],
                span: Union[int, np.ndarray] = None,
                ewm_lambda: Union[float, np.ndarray] = 0.94,
                init_value: Union[float, np.ndarray, None] = None,
                init_type: InitType = InitType.X0,
                is_unit_vol_scaling: bool = False,
                nan_backfill: NanBackfill = NanBackfill.FFILL,
                is_exlude_weekends: bool = False
                ) -> Union[pd.DataFrame, pd.Series, np.ndarray]:
    """
    ewm for pandas or series
    data dimension = t*n
    use gen data of pandas and np and call ewm_np with numa
    """
    if is_exlude_weekends:  # exclude weekends: need to adjust af accordingly
        if isinstance(data, pd.DataFrame) or isinstance(data, pd.Series):
            if isinstance(data.index, pd.DatetimeIndex):
                data = data.where(data.index.dayofweek < 5, other=np.nan)

    a = npo.to_finite_np(data=data, fill_value=np.nan)

    if init_value is None:
        init_value = set_init_dim1(data=a, init_type=init_type)

    # important for numba to have uniform data
    if isinstance(data, pd.Series) or (isinstance(data, np.ndarray) and data.ndim == 1):
        ewm_lambda = float(ewm_lambda)
        if isinstance(init_value, np.ndarray):
            init_value = float(init_value)

    ewm = ewm_recursion(a=a,
                        span=span,
                        ewm_lambda=ewm_lambda,
                        init_value=init_value,
                        is_unit_vol_scaling=is_unit_vol_scaling,
                        nan_backfill=nan_backfill)

    if isinstance(data, pd.DataFrame):  # return of data type
        ewm = pd.DataFrame(data=ewm, index=data.index, columns=data.columns)

    elif isinstance(data, pd.Series):  # return of data type
        ewm = pd.Series(data=ewm, index=data.index, name=data.name)

    return ewm


def compute_ewm_vol(data: Union[pd.DataFrame, pd.Series, np.ndarray],
                    ewm_lambda: Union[float, np.ndarray] = 0.94,
                    span: Optional[Union[float, np.ndarray]] = None,
                    mean_adj_type: MeanAdjType = MeanAdjType.NONE,
                    init_type: InitType = InitType.X0,
                    init_value: Optional[Union[float, np.ndarray]] = None,
                    apply_sqrt: bool = True,
                    annualize: bool = False,
                    af: Optional[float] = None,
                    nan_backfill: NanBackfill = NanBackfill.FFILL,
                    is_exlude_weekends: bool = False
                    ) -> Union[pd.DataFrame, pd.Series, np.ndarray]:
    """
    implementation of ewm recursion for variance/volatility computation
    """
    if is_exlude_weekends:  # exclude weekends: need to adjust af accordingly
        if isinstance(data, pd.DataFrame) or isinstance(data, pd.Series):
            if isinstance(data.index, pd.DatetimeIndex):
                data = data.where(data.index.dayofweek < 5, other=np.nan)

    a = npo.to_finite_np(data=data, fill_value=np.nan)

    if span is not None:
        ewm_lambda = 1.0 - 2.0 / (span + 1.0)

    if mean_adj_type != MeanAdjType.NONE:
        a = compute_rolling_mean_adj(data=a,
                                     mean_adj_type=mean_adj_type,
                                     ewm_lambda=ewm_lambda,
                                     init_type=init_type,
                                     nan_backfill=nan_backfill)

    # initial conditions
    a = np.square(a)
    if init_value is None:
        init_value = set_init_dim1(data=a, init_type=init_type)

    if isinstance(data, pd.Series) or (isinstance(data, np.ndarray) and data.ndim == 1):
        ewm_lambda = float(ewm_lambda)
        if isinstance(init_value, np.ndarray):
            init_value = float(init_value)

    ewm = ewm_recursion(a=a, ewm_lambda=ewm_lambda, init_value=init_value, nan_backfill=nan_backfill)

    if annualize or af is not None:
        if af is None:
            if isinstance(data, pd.DataFrame) or isinstance(data, pd.Series):
                af = da.infer_an_from_data(data=data)
            else:
                warnings.warn(f"in compute_ewm  annualization_factor for np array default is 1")
                af = 1.0
        ewm = af * ewm

    if apply_sqrt:
        ewm = np.sqrt(ewm)

    if isinstance(data, pd.DataFrame):
        ewm = pd.DataFrame(data=ewm, index=data.index, columns=data.columns)
    elif isinstance(data, pd.Series):
        ewm = pd.Series(data=ewm, index=data.index, name=data.name)

    return ewm


def compute_roll_mean(data: Union[pd.DataFrame, pd.Series, np.ndarray],
                      mean_adj_type: MeanAdjType = MeanAdjType.EWMA,
                      span: Union[int, np.ndarray] = None,
                      ewm_lambda: Union[float, np.ndarray] = 0.94,
                      init_value: Union[float, np.ndarray] = None,
                      nan_backfill: NanBackfill = NanBackfill.FFILL
                      ) -> Union[pd.DataFrame, pd.Series, np.ndarray]:

    """
    compute rolling mean by columns
    the output has the same dimension as input
    """
    if not (isinstance(data, pd.DataFrame) or isinstance(data, pd.Series) or isinstance(data, np.ndarray)):
        raise TypeError(f"unsupported type {type(data)}")

    if mean_adj_type == MeanAdjType.NONE:
        if isinstance(data, np.ndarray):
            mean = np.zeros_like(data)
        else:
            mean_nd = np.zeros_like(data.to_numpy())
            if isinstance(data, pd.DataFrame):
                mean = pd.DataFrame(data=mean_nd, index=data.index, columns=data.columns)
            else:
                mean = pd.Series(data=mean_nd, index=data.index, name=data.name)

    elif mean_adj_type == MeanAdjType.INSAMPLE:
        if isinstance(data, np.ndarray):
            mean = np.mean(data, axis=0, keepdims=True)
        else:
            mean_nd = np.mean(data.to_numpy(), axis=0, keepdims=True)
            if isinstance(data, pd.DataFrame):
                mean = pd.DataFrame(data=mean_nd, index=data.index, columns=data.columns)
            else:
                mean = pd.Series(data=mean_nd, index=data.index, name=data.name)

    elif mean_adj_type == MeanAdjType.EXPANDING:  # use pandas core
        if isinstance(data, pd.DataFrame) or isinstance(data, pd.Series):
            x = data
        else:  # convert to pandas
            x = pd.DataFrame(data=data)
        mean = x.expanding(min_periods=1, axis=0).mean()  # apply pandas expanding
        if isinstance(data, np.ndarray):  # return of np.ndarray data type
            mean = mean.to_numpy()

    elif mean_adj_type == MeanAdjType.EWMA:
        mean = compute_ewm(data=data,
                           span=span,
                           ewm_lambda=ewm_lambda,
                           init_value=init_value,
                           nan_backfill=nan_backfill)
    else:
        raise TypeError(f"mean_adj_type={mean_adj_type} is not implemented")

    return mean


def compute_rolling_mean_adj(data: Union[pd.DataFrame, pd.Series, np.ndarray],
                             mean_adj_type: MeanAdjType = MeanAdjType.EWMA,
                             span: Union[int, np.ndarray] = None,
                             ewm_lambda: Union[float, np.ndarray] = 0.94,
                             init_type: InitType = InitType.MEAN,
                             init_value: Union[float, np.ndarray, None] = None,
                             nan_backfill: NanBackfill = NanBackfill.FFILL
                             ) -> Union[pd.DataFrame, pd.Series, np.ndarray]:

    if mean_adj_type == MeanAdjType.NONE:
        x_mean = data
    else:
        if init_value is None:
            init_value = set_init_dim1(data=data, init_type=init_type)

        mean = compute_roll_mean(data=data,
                                 mean_adj_type=mean_adj_type,
                                 span=span,
                                 ewm_lambda=ewm_lambda,
                                 init_value=init_value,
                                 nan_backfill=nan_backfill)
        x_mean = data - mean

    return x_mean


def compute_ewm_cross_xy(x_data: Union[pd.DataFrame, pd.Series, np.ndarray],
                         y_data: Union[pd.DataFrame, pd.Series, np.ndarray],
                         span: Union[int, np.ndarray] = None,
                         ewm_lambda: Union[float, np.ndarray] = 0.94,
                         cross_xy_type: CrossXyType = CrossXyType.COVAR,
                         mean_adj_type: MeanAdjType = MeanAdjType.NONE,
                         init_type: InitType = InitType.ZERO,
                         var_init_type: InitType = InitType.MEAN,  # to avoid overflows
                         nan_backfill: NanBackfill = NanBackfill.FFILL
                         ) -> Union[pd.DataFrame, pd.Series, np.ndarray]:
    """
    compute cross ewm for 1-d arrays x and y
    cross_xy[t] = (1-lambda)*x[t]*y[t] + lambda*cross_xy[t-1]

    three supported cases:
    1: x and y are pd.DataFrame with same dimensions: z = x*y
    2: x is pd.Series, y is pd.DataFrame: ml y=betas*x -> output pandas.dim = pandas.dim y
    3: both pd.Series -> output series.dim = series.dim y
    4: both np.nd arrays with same dimension
    """

    # 1 - adjust by mean
    if mean_adj_type != MeanAdjType.NONE:

        x_data = compute_rolling_mean_adj(data=x_data,
                                          mean_adj_type=mean_adj_type,
                                          span=span,
                                          ewm_lambda=ewm_lambda,
                                          init_type=init_type,
                                          nan_backfill=nan_backfill)

        y_data = compute_rolling_mean_adj(data=y_data,
                                          mean_adj_type=mean_adj_type,
                                          span=span,
                                          ewm_lambda=ewm_lambda,
                                          init_type=init_type,
                                          nan_backfill=nan_backfill)

    # 2  take gen arrays and convert to ndarray to use with numbas
    if isinstance(x_data, pd.DataFrame) and isinstance(y_data, pd.DataFrame):
        # should be same dimensions
        x = npo.to_finite_np(data=x_data, fill_value=np.nan)
        y = npo.to_finite_np(data=y_data, fill_value=np.nan)
        xy = np.multiply(x, y)

    elif isinstance(x_data, pd.DataFrame) and isinstance(y_data, pd.Series):
        xy = pd.concat([x_data, y_data], axis=1, join='inner')

        # it will work even if x_data.name is in y_data.columns
        x = npo.to_finite_np(data=xy.iloc[:, 0], fill_value=np.nan)
        y = npo.to_finite_np(data=xy.iloc[:, 1:], fill_value=np.nan)

        # x is array: tile by rows and transpose
        xn = np.transpose(np.tile(x, (len(y_data.columns), 1)))
        xy = np.multiply(xn, y)

    elif isinstance(x_data, pd.Series) and isinstance(y_data, pd.Series):
        xy = pd.concat([x_data, y_data], axis=1, join='inner')
        x = npo.to_finite_np(data=xy.iloc[:, 0], fill_value=np.nan)
        y = npo.to_finite_np(data=xy.iloc[:, 1], fill_value=np.nan)
        xy = np.multiply(x, y)

    elif isinstance(x_data, np.ndarray) and isinstance(y_data, np.ndarray):
        if x_data.shape[1] != y_data.shape[1]:
            raise TypeError(f"ndarray data must have same number of column")
        if x_data.shape[0] != y_data.shape[0]:
            raise TypeError(f"ndarray data must have same number of rows")

        x = x_data
        y = y_data
        xy = np.multiply(x_data, y_data)

    else:
        raise TypeError(f"{type(x_data)}, {type(y_data)}  should be of the same type")

    init_value_xy = set_init_dim1(data=xy, init_type=init_type)
    xy_covar = ewm_recursion(a=xy,
                             span=span,
                             ewm_lambda=ewm_lambda,
                             init_value=init_value_xy,
                             nan_backfill=nan_backfill)

    if cross_xy_type == CrossXyType.COVAR:
        cross_xy = xy_covar

    elif cross_xy_type == CrossXyType.BETA:
        x2 = np.square(x)
        init_value_x2 = set_init_dim1(data=x2, init_type=var_init_type)
        x_var = ewm_recursion(a=x2,
                              span=span,
                              ewm_lambda=ewm_lambda,
                              init_value=init_value_x2,
                              nan_backfill=nan_backfill)
        divisor = x_var
        cross_xy = np.divide(xy_covar, divisor, where=np.isclose(divisor, 0.0) == False)

    elif cross_xy_type == CrossXyType.CORR:
        x2 = np.square(x)
        init_value_x2 = set_init_dim1(data=x2, init_type=var_init_type)
        x_var = ewm_recursion(a=x2, span=span, ewm_lambda=ewm_lambda, init_value=init_value_x2, nan_backfill=nan_backfill)
        y2 = np.square(y)
        init_value_y2 = set_init_dim1(data=y2, init_type=var_init_type)
        y_var = ewm_recursion(a=y2, span=span, ewm_lambda=ewm_lambda, init_value=init_value_y2, nan_backfill=nan_backfill)
        divisor = np.sqrt(np.multiply(x_var, y_var))
        cross_xy = np.divide(xy_covar, divisor, where=np.isclose(divisor, 0.0) == False)
    else:
        raise TypeError(f"unknown cross_xy_type = {cross_xy_type}")

    if isinstance(y_data, pd.Series):
        cross_xy = pd.Series(data=cross_xy, index=y_data.index, name=y_data.name)

    elif isinstance(y_data, pd.DataFrame):
        cross_xy = pd.DataFrame(data=cross_xy, index=y_data.index, columns=y_data.columns)

    return cross_xy


def compute_ewm_beta_resid(x_data: pd.DataFrame,
                           y_data: pd.DataFrame,
                           span: Union[int, np.ndarray] = None,
                           ewm_lambda: Union[float, np.ndarray] = 0.94,
                           mean_adj_type: MeanAdjType = MeanAdjType.NONE,
                           init_type: InitType = InitType.MEAN,
                           annualize: bool = False,
                           nan_backfill: NanBackfill = NanBackfill.FFILL
                           ) -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]:

    """
    compute cross ewm for 1-d arrays x and y
    cross_xy[t] = (1-lambda)*x[t]*y[t] + lambda*cross_xy[t-1]
    returns betas and annualized vars of resid and beta
    """
    # 1 - adjust by mean
    if mean_adj_type != MeanAdjType.NONE:
        x_data = compute_rolling_mean_adj(data=x_data,
                                          mean_adj_type=mean_adj_type,
                                          span=span,
                                          ewm_lambda=ewm_lambda,
                                          init_type=init_type,
                                          nan_backfill=nan_backfill)

        y_data = compute_rolling_mean_adj(data=y_data,
                                          mean_adj_type=mean_adj_type,
                                          span=span,
                                          ewm_lambda=ewm_lambda,
                                          init_type=init_type,
                                          nan_backfill=nan_backfill)

    # 2  take gen arrays and convert to ndarray to use with numbdas
    if isinstance(x_data, pd.DataFrame) and isinstance(y_data, pd.DataFrame):
        # should be same dimensions
        x = npo.to_finite_np(data=x_data, fill_value=np.nan)
        y = npo.to_finite_np(data=y_data, fill_value=np.nan)
        xy = np.multiply(x, y)
    else:
        raise TypeError(f"in compute_ewm_beta_resid: x and y should be type pd.DataFrame")

    initial_value_xy = set_init_dim1(data=xy, init_type=init_type)
    xy_covar = ewm_recursion(a=xy, ewm_lambda=ewm_lambda, init_value=initial_value_xy)

    x2 = np.square(x)
    initial_value_x2 = set_init_dim1(data=x2, init_type=init_type)
    x_var = ewm_recursion(a=x2, ewm_lambda=ewm_lambda, init_value=initial_value_x2)

    beta_xy = np.divide(xy_covar, x_var, where=np.isclose(x_var, 0.0)==False)

    resid = y - beta_xy * npo.np_array_to_n_column_array(a=x, n_col=beta_xy.shape[1]) # extend x by columns to match columns of beta_xy
    resid2 = np.square(resid)
    iv_resid2 = set_init_dim1(data=resid2, init_type=init_type)
    resid_var = ewm_recursion(a=resid2, ewm_lambda=ewm_lambda, init_value=iv_resid2)

    if annualize:
        if isinstance(x_data, pd.DataFrame) or isinstance(x_data, pd.Series):
            an = da.infer_an_from_data(data=x_data)
        else:
            warnings.warn(f"in compute_ewm  annualization_factor for np array default is 1")
            an = 1.0
        resid_var = an * resid_var
        x_var = an * x_var

    beta_xy = pd.DataFrame(data=beta_xy, index=y_data.index, columns=y_data.columns)
    resid_var = pd.DataFrame(data=resid_var, index=y_data.index, columns=y_data.columns)
    # resid_var = pd.DataFrame(data=an * resid2, index=y_data.index, columns=y_data.columns).expanding().var()
    x_var = pd.DataFrame(data=x_var, index=y_data.index)

    return beta_xy, x_var, resid_var


def compute_ewm_alpha_r2(y_data: pd.DataFrame,
                         y_prediction: pd.DataFrame,
                         span: Union[int, np.ndarray] = None,
                         ewm_lambda: Union[float, np.ndarray] = 0.94,
                         nan_backfill: NanBackfill = NanBackfill.FFILL
                         ) -> Tuple[pd.DataFrame, pd.DataFrame]:
    """
    """
    # 1 - adjust by mean
    resid = y_data - y_prediction
    ewm_alpha = compute_ewm(data=resid, span=span, ewm_lambda=ewm_lambda, nan_backfill=nan_backfill)
    resid0 = resid.subtract(ewm_alpha)
    resid_var = ewm_recursion(a=np.square(resid0.to_numpy()), span=span, ewm_lambda=ewm_lambda,
                              init_value=np.zeros(len(y_data.columns)), nan_backfill=nan_backfill)

    y_var0 = y_data.subtract(compute_ewm(data=y_data, span=span, ewm_lambda=ewm_lambda, nan_backfill=nan_backfill))
    y_var = ewm_recursion(a=np.square(y_var0.to_numpy()), span=span, ewm_lambda=ewm_lambda,
                          init_value=np.zeros(len(y_data.columns)), nan_backfill=nan_backfill)

    ewm_r2 = 1.0 - np.divide(resid_var, y_var, where=np.greater(y_var, 0.0))
    ewm_r2 = np.clip(ewm_r2, a_min=0.0, a_max=1.0)
    ewm_r2 = pd.DataFrame(data=ewm_r2, index=y_data.index, columns=y_data.columns)

    return ewm_alpha, ewm_r2


def compute_ewm_sharpe(returns: pd.DataFrame,
                       span: Union[int, np.ndarray] = 260,
                       norm_type: int = 1,
                       initial_sharpes: np.ndarray = None
                       ) -> pd.DataFrame:

    x = npo.to_finite_np(data=returns, fill_value=0.0)
    an = da.infer_an_from_data(data=returns)
    san = np.sqrt(an)
    if initial_sharpes is not None:
        initial_vol = 0.1
        initial_mean = initial_vol * initial_sharpes / np.square(san)
        initial_var = np.square(initial_vol / san) * np.ones(len(returns.columns))
    else:
        initial_mean = np.zeros(len(returns.columns))
        initial_var = np.zeros(len(returns.columns))

    if norm_type == 0:
        ewm_mean = ewm_recursion(a=x,
                                 span=span,
                                 init_value=initial_mean,
                                 nan_backfill=NanBackfill.ZERO_FILL)
        sharpe = pd.DataFrame(data=an*ewm_mean, index=returns.index, columns=returns.columns)

    elif norm_type == 1 or norm_type == 2:

        ewm_mean = ewm_recursion(a=x,
                                 span=span,
                                 init_value=initial_mean,
                                 nan_backfill=NanBackfill.ZERO_FILL)
        if norm_type == 2:
            v = np.square(x-ewm_mean)
        else:
            v = np.square(x)

        ewm_var = ewm_recursion(a=v, span=span,
                                init_value=initial_var,
                                nan_backfill=NanBackfill.ZERO_FILL)
        ewm_vol = np.sqrt(ewm_var)
        sharpe = pd.DataFrame(data=san * np.divide(ewm_mean, ewm_vol, where=np.greater(ewm_vol, 0.0)),
                              index=returns.index,
                              columns=returns.columns)
    else:
        raise ValueError(f"norm_type={norm_type} not implemented")

    return sharpe


def compute_ewm_sharpe_from_prices(prices: pd.DataFrame,
                                   freq: str = 'QE',
                                   span: int = 40,
                                   initial_sharpes: np.ndarray = None,
                                   norm_type: int = 2
                                   ) -> pd.DataFrame:

    prices = prices.asfreq(freq=freq, method='ffill')
    returns = np.log(prices.divide(prices.shift(1)))
    sharpe = compute_ewm_sharpe(returns=returns,
                                span=span,
                                initial_sharpes=initial_sharpes,
                                norm_type=norm_type)

    return sharpe


def compute_ewm_std1_norm(data: Union[pd.DataFrame, pd.Series],
                          span: Union[int, np.ndarray] = 260,
                          mean_adj_type: MeanAdjType = MeanAdjType.EWMA,
                          is_demean: bool = True,
                          is_nans_to_zero: bool = True
                          ) -> Union[pd.DataFrame, pd.Series]:
    """
    given time serise var X_t
    compute Y_t = ewm(X_t / ewm_vol(X_t))
    expected std(Y_t) = 1
    """

    data_np = npo.to_finite_np(data=data, fill_value=np.nan)

    if is_demean:
        x_mean = compute_roll_mean(data=data_np, mean_adj_type=mean_adj_type, span=span)
        x_demean = data_np-x_mean
    else:
        x_demean = data_np

    ewm_var = compute_ewm(data=np.square(x_demean),
                          span=span,
                          is_unit_vol_scaling=False)
    ewm_vol = np.sqrt(ewm_var)

    x_std1_norm = npo.to_finite_ratio(x=x_demean, y=ewm_vol, fill_value=np.nan)
    ewm_std1_norm = compute_ewm(data=x_std1_norm, span=span, is_unit_vol_scaling=True)

    if isinstance(data, pd.Series):
        ewm_std1_norm = pd.Series(data=ewm_std1_norm, index=data.index, name=data.name)
    else:
        ewm_std1_norm = pd.DataFrame(data=ewm_std1_norm, index=data.index, columns=data.columns)

    if is_nans_to_zero:
        ewm_std1_norm = ewm_std1_norm.fillna(0.0)

    return ewm_std1_norm


# @njit
def ewm_vol_assymetric_np(returns: np.ndarray,
                        ewm_lambda: Union[float, np.ndarray] = 0.94,
                        annualization_factor: float = 1.0
                        ) ->Tuple[np.ndarray, np.ndarray]:
    """
    applies strictly to numpy arrays with utilization of numbda
    data: numpy with dimension = t*n
    ewm_lambda: float or ndarray of dimension n
    init_value: initial value of dimension n
    """
    ewm_lambda_1 = 1.0 - ewm_lambda

    # initialize all
    ewm_m, ewm_p = np.zeros_like(returns), np.zeros_like(returns)
    returns2 = np.square(returns)
    return2_m = np.where(np.less(returns, 0.0), returns2, np.nan)
    return2_p = np.where(np.greater(returns, 0.0), returns2, np.nan)
    ewm_m[0], ewm_p[0] = np.mean(return2_m[~np.isnan(return2_m)], axis=0), np.mean(return2_p[~np.isnan(return2_p)], axis=0)
    ewm_m0 = ewm_m[0]
    ewm_p0 = ewm_p[0]

    nt = returns.shape[0]
    for t in np.arange(1, nt):  # for x_t in x[1:]: # got by rows in x

        ewm_m1_ = ewm_lambda * ewm_m0 + ewm_lambda_1 * return2_m[t]
        ewm_m1 = np.where(np.isnan(return2_m[t]), ewm_m0, ewm_m1_)

        ewm_p1_ = ewm_lambda * ewm_p0 + ewm_lambda_1 * return2_p[t]
        ewm_p1 = np.where(np.isnan(return2_p[t]), ewm_p0, ewm_p1_)

        ewm_m[t] = ewm_m0 = ewm_m1
        ewm_p[t] = ewm_p0 = ewm_p1

    ewm_m = np.sqrt(annualization_factor*ewm_m)
    ewm_p = np.sqrt(annualization_factor*ewm_p)

    return ewm_m, ewm_p


def ewm_vol_assymetric(returns: Union[pd.Series, pd.DataFrame],
                        ewm_lambda: Union[float, np.ndarray] = 0.94,
                        annualization_factor: float = 1.0
                        ) -> Tuple[Union[pd.Series, pd.DataFrame], Union[pd.Series, pd.DataFrame]]:

    ewm_m, ewm_p = ewm_vol_assymetric_np(returns=returns.to_numpy(),
                                         ewm_lambda=ewm_lambda,
                                         annualization_factor=annualization_factor)
    if isinstance(returns, pd.DataFrame):
        ewm_m = pd.DataFrame(ewm_m, index=returns.index, columns=returns.columns)
        ewm_p = pd.DataFrame(ewm_p, index=returns.index, columns=returns.columns)
    elif isinstance(returns, pd.Series):
        ewm_m = pd.Series(ewm_m, index=returns.index, name=returns.name)
        ewm_p = pd.Series(ewm_p, index=returns.index, name=returns.name)
    else:
        raise TypeError
    return ewm_m, ewm_p



"""
analytics for fx data
"""
import pandas as pd
import numpy as np
from typing import Union, Dict


def get_aligned_fx_spots(prices: pd.DataFrame,
                         asset_ccy_map: Union[pd.Series, Dict],
                         fx_prices: pd.DataFrame,
                         quote_currency: str = 'USD'
                         ) -> pd.DataFrame:
    """
    get fx currency for price_data columns = instrument ticker
    universe_local_ccy is map {instrument ticker: fx_rate_ccy}
    fx_prices is fx prices columns = fx_rate_ccy
    """
    # first backfill and the bbfill so prices will have corresponding fx spots data
    fx_prices = fx_prices.copy().ffill().bfill()
    fx_prices[quote_currency] = 1.0

    fx_spots = {}
    for asset, ccy in asset_ccy_map.items():
        fx_spots[asset] = fx_prices[ccy]
    fx_spots = pd.DataFrame.from_dict(fx_spots, orient='columns')
    fx_spots = fx_spots.where(pd.isna(prices) == False)
    return fx_spots


def compute_futures_fx_adjusted_returns(prices: pd.DataFrame,
                                        fx_spots: pd.DataFrame,
                                        periods: int = 1,
                                        is_log_returns: bool = False
                                        ) -> pd.DataFrame:
    """
    futures returns adjusted for fx rate change
    fx_return affects only the future return
    """
    price_return = prices / prices.shift(periods=periods) - 1.0
    fx_return = fx_spots / fx_spots.shift(periods=periods) - 1.0
    returns = np.log(1.0 + price_return + price_return * fx_return)
    if not is_log_returns:
        returns = np.expm1(returns)
    return returns


def compute_cash_fx_adjusted_returns(prices: pd.DataFrame,
                                     fx_spots: pd.DataFrame,
                                     periods: int = 1,
                                     is_log_returns: bool = False
                                     ) -> pd.DataFrame:
    """
    cash returns adjusted for fx rate change
    fx_return affects notional + return
    """
    price_return = prices / prices.shift(periods=periods) - 1.0
    fx_return = fx_spots / fx_spots.shift(periods=periods) - 1.0
    returns = np.log(1.0 + fx_return + price_return + price_return * fx_return)
    if not is_log_returns:
        returns = np.expm1(returns)
    return returns



import numpy as np
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt
from enum import Enum
from typing import Union
import yfinance as yf
import qis


def fetch_hourly_data(ticker: str = 'BTC-USD') -> pd.DataFrame:
    """
    yf reports timestamps of bars at the start of the period: we shift it to the end of the period
    """
    asset = yf.Ticker(ticker)
    ohlc_data = asset.history(period="730d", interval="1h")
    ohlc_data.index = [t + pd.Timedelta(minutes=60) for t in ohlc_data.index]
    ohlc_data = ohlc_data.rename({'Open': 'open', 'High': 'high', 'Low': 'low', 'Close': 'close'}, axis=1)
    ohlc_data.index = ohlc_data.index.tz_convert('UTC')
    ohlc_data.index.name = 'timestamp'
    return ohlc_data


def compute_vols(prices: pd.Series,
                 span: int = 5*24
                 ) -> pd.DataFrame:
    """
    compute rolling ewma with and without weekends
    need to adjust annualisation factor
    """
    returns = np.log(prices).diff(1)
    init_value = np.nanvar(returns, axis=0)  # set initial value to average variance
    vol = qis.compute_ewm_vol(data=returns, span=span, af=365*24, init_value=init_value)
    vol1 = qis.compute_ewm_vol(data=returns, span=span, af=260*24, init_value=init_value, is_exlude_weekends=True)
    vols = pd.concat([vol.rename('including weekends'),
                      vol1.rename('excluding weekends')], axis=1)
    return vols


class UnitTests(Enum):
    VOLS = 1


def run_unit_test(unit_test: UnitTests):

    if unit_test == UnitTests.VOLS:
        prices = fetch_hourly_data(ticker='BTC-USD')['close']
        vol = compute_vols(prices=prices)
        with sns.axes_style("darkgrid"):
            fig, ax = plt.subplots(1, 1, figsize=(10, 7))
            qis.plot_time_series(df=vol, var_format='{:,.2%}', ax=ax)

    plt.show()


if __name__ == '__main__':

    unit_test = UnitTests.VOLS

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)




# packages
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from enum import Enum
from typing import Optional, Union

# qis
import qis.utils.dates as da
import qis.utils.np_ops as npn

import qis.perfstats.returns as ret
from qis.perfstats.config import PerfParams, ReturnTypes
from qis.perfstats.regime_classifier import BenchmarkReturnsQuantileRegimeSpecs

import qis.plots.time_series as pts
import qis.plots.utils as put
import qis.plots.heatmap as phe
from qis.plots.derived.regime_data import add_bnb_regime_shadows

import qis.models.linear.corr_cov_matrix as ccm
import qis.models.linear.ewm as ewm
from qis.models.linear.corr_cov_matrix import CorrMatrixOutput
from qis.models.linear.ewm import InitType


def plot_returns_corr_table(prices: pd.DataFrame,
                            var_format: str = '{:.0%}',
                            freq: Optional[str] = None,
                            cmap: str = 'PiYG',
                            return_type: ReturnTypes = ReturnTypes.RELATIVE,
                            is_fig_out: bool = True,
                            ax: plt.Subplot = None,
                            **kwargs
                            ) -> Union[Optional[plt.Figure], pd.DataFrame]:
    """
    plot corr table of prices
    """
    returns = ret.to_returns(prices=prices, return_type=return_type, freq=freq)
    corr = ccm.compute_masked_covar_corr(data=returns, is_covar=False)
    if is_fig_out:
        return phe.plot_heatmap(df=corr,
                               var_format=var_format,
                               cmap=cmap,
                               ax=ax,
                               **kwargs)
    else:
        return corr


def plot_returns_ewm_corr_table(prices: pd.DataFrame,
                                span: Union[int, np.ndarray] = None,
                                ewm_lambda: float = 0.94,
                                var_format: str = '{:.0%}',
                                return_type: ReturnTypes = ReturnTypes.LOG,
                                init_type: InitType = InitType.ZERO,
                                freq: Optional[str] = None,
                                cmap: str = 'PiYG',
                                is_last: bool = True,
                                ax: plt.Subplot = None,
                                **kwargs
                                ) -> plt.Figure:
    """
     plot corr table of rices
    """
    returns = ret.to_returns(prices=prices, return_type=return_type, freq=freq)
    init_value = ewm.set_init_dim2(data=returns.to_numpy(), init_type=init_type)
    corr = ewm.compute_ewm_covar_tensor(a=returns.to_numpy(),
                                        span=span,
                                        ewm_lambda=ewm_lambda,
                                        covar0=init_value,
                                        is_corr=True)
    if is_last:
        ar = corr[-1]
    else:
        ar = npn.tensor_mean(corr)

    df = pd.DataFrame(ar, index=prices.columns, columns=prices.columns)
    fig = phe.plot_heatmap(df=df,
                           var_format=var_format,
                           cmap=cmap,
                           ax=ax,
                           **kwargs)
    return fig


def plot_returns_corr_matrix_time_series(prices: pd.DataFrame,
                                         corr_matrix_output: CorrMatrixOutput = CorrMatrixOutput.FULL,
                                         return_type: ReturnTypes = ReturnTypes.LOG,
                                         time_period: da.TimePeriod = None,
                                         freq: Optional[str] = None,
                                         span: Union[int, np.ndarray] = None,
                                         ewm_lambda: float = 0.97,
                                         init_type: InitType = InitType.X0,
                                         var_format: str = '{:.0%}',
                                         legend_stats: pts.LegendStats = pts.LegendStats.AVG_LAST,
                                         trend_line: put.TrendLine = put.TrendLine.AVERAGE,
                                         regime_benchmark_str: str = None,
                                         regime_params: BenchmarkReturnsQuantileRegimeSpecs = None,
                                         perf_params: PerfParams = None,
                                         ax: plt.Subplot = None,
                                         **kwargs
                                         ) -> None:

    """
    plot time series of returns correlations infered from prices
    """
    returns = ret.to_returns(prices=prices, return_type=return_type, freq=freq)
    corr_pandas = ccm.compute_ewm_corr_df(df=returns,
                                          corr_matrix_output=corr_matrix_output,
                                          span=span,
                                          ewm_lambda=ewm_lambda,
                                          init_type=init_type)
    if time_period is not None:
        corr_pandas = time_period.locate(corr_pandas)

    pts.plot_time_series(df=corr_pandas,
                         legend_stats=legend_stats,
                         trend_line=trend_line,
                         var_format=var_format,
                         ax=ax,
                         **kwargs)
    if regime_benchmark_str is not None:
        if regime_benchmark_str in prices.columns:
            pivot_prices = prices[regime_benchmark_str].reindex(index=corr_pandas.index, method='ffill')
        else:
            raise KeyError(f"{regime_benchmark_str} not in {prices.columns}")

        if regime_benchmark_str is not None and regime_params is not None:
            add_bnb_regime_shadows(ax=ax,
                                   pivot_prices=pivot_prices,
                                   benchmark=regime_benchmark_str,
                                   regime_params=regime_params,
                                   perf_params=perf_params)


class UnitTests(Enum):
    CORR_TABLE = 1
    CORR_MATRIX = 2
    EWMA_CORR = 3


def run_unit_test(unit_test: UnitTests):

    from qis.test_data import load_etf_data
    prices = load_etf_data().dropna()

    if unit_test == UnitTests.CORR_TABLE:
        plot_returns_corr_table(prices=prices)

    elif unit_test == UnitTests.CORR_MATRIX:
        fig, ax = plt.subplots(1, 1, figsize=(10, 10), constrained_layout=True)
        plot_returns_corr_matrix_time_series(prices=prices, regime_benchmark_str='SPY', ax=ax)

    elif unit_test == UnitTests.EWMA_CORR:
        plot_returns_ewm_corr_table(prices=prices.iloc[:, :5])

    plt.show()


if __name__ == '__main__':

    unit_test = UnitTests.CORR_MATRIX

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)



"""
module for computing rolling performance stats
"""
# packages
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from scipy.stats import skew
from enum import Enum
from typing import Union, Optional

# qis
from qis.perfstats import returns as ret
from qis.utils import dates as da


class RollingPerfStat(str, Enum):
    TOTAL_RETURNS = 'Total returns'
    PA_RETURNS = 'Pa returns'
    VOL = 'Volatility'
    SHARPE = 'Sharp ratio'
    SKEW = 'Skeweness'


def compute_rolling_perf_stat(prices: Union[pd.DataFrame, pd.Series],
                              rolling_perf_stat: RollingPerfStat = RollingPerfStat.TOTAL_RETURNS,
                              roll_freq: Optional[str] = None,
                              roll_periods: int = 260
                              ) -> Union[pd.DataFrame, pd.Series]:
    """
    compute rolling performance
    default stats is 1y rolling for prices with roll_freq = 'B'
    for monthly returns anf 5y rolling use roll_freq='ME' and roll_periods=5*12
    """
    if rolling_perf_stat == RollingPerfStat.TOTAL_RETURNS:
        perf_stat = compute_rolling_returns(prices=prices, roll_periods=roll_periods)
    elif rolling_perf_stat == RollingPerfStat.PA_RETURNS:
        perf_stat = compute_rolling_pa_returns(prices=prices, roll_periods=roll_periods)
    elif rolling_perf_stat == RollingPerfStat.VOL:
        perf_stat = compute_rolling_vols(prices=prices, roll_freq=roll_freq, roll_periods=roll_periods)
    elif rolling_perf_stat == RollingPerfStat.SHARPE:
        perf_stat = compute_rolling_sharpes(prices=prices, roll_freq=roll_freq, roll_periods=roll_periods)
    elif rolling_perf_stat == RollingPerfStat.SKEW:
        perf_stat = compute_rolling_skew(prices=prices, roll_freq=roll_freq, roll_periods=roll_periods)
    else:
        raise NotImplementedError(f"rolling_perf_stats")
    return perf_stat


def compute_rolling_returns(prices: Union[pd.DataFrame, pd.Series],
                            roll_periods: int = 260
                            ) -> Union[pd.DataFrame, pd.Series]:
    """
    compute rolling returns
    """
    returns = prices.divide(prices.shift(periods=roll_periods)) - 1.0
    return returns


def compute_rolling_pa_returns(prices: Union[pd.DataFrame, pd.Series],
                               roll_periods: int = 260
                               ) -> Union[pd.DataFrame, pd.Series]:
    """
    compute rolling pa returns
    """
    pa_returns = prices.rolling(roll_periods).apply(lambda x: ret.compute_pa_return(x))
    return pa_returns


def compute_rolling_vols(prices: Union[pd.Series, pd.DataFrame],
                        roll_freq: Optional[str] = None,
                        roll_periods: int = 260
                        ) -> Union[pd.Series, pd.DataFrame]:
    log_returns = ret.to_returns(prices=prices, freq=roll_freq, is_log_returns=True, drop_first=False)
    saf = np.sqrt(da.infer_an_from_data(data=log_returns))
    vols = saf * log_returns.rolling(roll_periods).apply(lambda x: np.nanstd(x, ddof=1))
    return vols


def compute_rolling_sharpes(prices: Union[pd.Series, pd.DataFrame],
                            roll_freq: Optional[str] = None,
                            roll_periods: int = 260
                            ) -> Union[pd.Series, pd.DataFrame]:
    log_returns = ret.to_returns(prices=prices, freq=roll_freq, is_log_returns=True, drop_first=False)
    saf = np.sqrt(da.infer_an_from_data(data=log_returns))
    sharpes = log_returns.rolling(roll_periods).apply(lambda x: compute_sharpe(x, saf=saf))
    return sharpes


def compute_sharpe(log_returns: Union[pd.Series, pd.DataFrame], saf: float = None) -> np.ndarray:
    if saf is None:
        saf = np.sqrt(da.infer_an_from_data(data=log_returns))
    mean = np.expm1(np.nanmean(log_returns.to_numpy()))
    vol = np.nanstd(log_returns.to_numpy(), ddof=1)
    if np.greater(vol, 0.0):
        sharpe = saf * mean / vol
    else:
        sharpe = np.nan
    return sharpe


def compute_rolling_skew(prices: Union[pd.Series, pd.DataFrame],
                         roll_freq: Optional[str] = None,
                         roll_periods: int = 120,
                         ) -> Union[pd.Series, pd.DataFrame]:
    log_returns = ret.to_returns(prices=prices, freq=roll_freq, is_log_returns=True, drop_first=False)
    skw = log_returns.rolling(roll_periods).apply(lambda x: compute_skew(x))
    return skw


def compute_skew(log_returns: Union[pd.Series, pd.DataFrame]) -> np.ndarray:
    skw = skew(log_returns.to_numpy(), axis=0, nan_policy='omit')
    return skw


class UnitTests(Enum):
    ROLLING_STATS = 1


def run_unit_test(unit_test: UnitTests):

    import qis.plots.time_series as pts
    from qis.test_data import load_etf_data
    prices = load_etf_data().dropna()

    if unit_test == UnitTests.ROLLING_STATS:

        for rolling_perf_stat in RollingPerfStat:
            stats = compute_rolling_perf_stat(prices=prices,
                                              rolling_perf_stat=rolling_perf_stat,
                                              roll_freq='W-WED',
                                              roll_periods=5*52)
            pts.plot_time_series(df=stats,
                                 var_format='{:.2f}',
                                 title=f"{rolling_perf_stat.name}")

    plt.show()


if __name__ == '__main__':

    unit_test = UnitTests.ROLLING_STATS

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)



from __future__ import annotations

import numpy as np
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt
from typing import Optional, Tuple
from enum import Enum

# qis
import qis.utils.df_cut as dfc
from qis.perfstats.config import RegimeData, PerfParams
from qis.perfstats.regime_classifier import RegimeClassifier, BenchmarkReturnsQuantileRegimeSpecs, \
    BenchmarkReturnsQuantilesRegime, VolQuantileRegimeSpecs, BenchmarkVolsQuantilesRegime, compute_bnb_regimes_pa_perf_table
import qis.plots.bars as pba
import qis.plots.boxplot as bxp


def plot_regime_data(regime_classifier: RegimeClassifier,
                     regime_data_to_plot: RegimeData = RegimeData.REGIME_SHARPE,
                     drop_sharpe_from_labels: bool = False,  # only leave the regime names
                     x_rotation: int = 0,
                     add_bar_values: bool = True,
                     title: Optional[str] = 'Conditional Excess Sharpe ratio',
                     var_format: str = '{:.1f}',
                     bbox_to_anchor: Optional[Tuple[float, float]] = (1.0, 0.95),
                     fontsize: int = 10,
                     is_top_totals: bool = True,
                     is_add_totals: bool = True,
                     is_use_vbar: bool = False,
                     reverse_columns: bool = True,
                     legend_loc: Optional[str] = 'upper right',
                     ax: plt.Subplot = None,
                     **kwargs
                     ) -> plt.Figure:

    regimes_pa_perf_table, regime_datas = regime_classifier.compute_regimes_pa_perf_table(**kwargs)
    data = regime_datas[regime_data_to_plot]
    data = data.dropna(how='all', axis=0)  # remove data wit all rows nans

    if drop_sharpe_from_labels:
        data.columns = [x.replace(' Sharpe', '') for x in data.columns]

    regime_colors = list(regime_classifier.get_regime_ids_colors().values())

    if is_add_totals:
        totals = np.nansum(data.to_numpy(), axis=1)
    else:
        totals = None

    if is_use_vbar:
        fig = pba.plot_vbars(df=data,
                             x_rotation=x_rotation,
                             colors=regime_colors,
                             var_format=var_format,
                             add_bar_values=add_bar_values,
                             title=title,
                             totals=totals,
                             fontsize=fontsize,
                             bbox_to_anchor=bbox_to_anchor,
                             xmin_shift=-0.05,
                             x_step=0.5,
                             ax=ax,
                             **kwargs)
    else:
        fig = pba.plot_bars(df=data,
                            x_rotation=x_rotation,
                            colors=regime_colors,
                            xvar_format=var_format,
                            add_bar_values=add_bar_values,
                            title=title,
                            totals=totals,
                            is_top_totals=is_top_totals,
                            fontsize=fontsize,
                            bbox_to_anchor=bbox_to_anchor,
                            legend_loc=legend_loc,
                            reverse_columns=reverse_columns,
                            ax=ax,
                            **kwargs)
    return fig


def plot_regime_boxplot(regime_classifier: RegimeClassifier,
                        prices: pd.DataFrame,
                        benchmark: str,
                        hue_var_name: str = 'Asset',
                        x_index_var_name: str = 'Regime',
                        ylabel: str = 'avg regime return',
                        x_rotation: int = 0,
                        legend_loc: Optional[str] = 'upper left',
                        yvar_format: str = '{:.0%}',
                        meanline: bool = True,
                        ax: plt.Subplot = None,
                        **kwargs
                        ) -> plt.Figure:

    df = regime_classifier.compute_sampled_returns_with_regime_id(prices=prices, benchmark=benchmark).iloc[1:, :]
    regime_ids_colors = regime_classifier.get_regime_ids_colors()
    df = df.set_index(regime_classifier.REGIME_COLUMN, drop=True) # boxplot b regime as hue
    df = dfc.sort_index_by_hue(df=df, hue_order=list(regime_ids_colors.keys()))
    fig = bxp.df_boxplot_by_hue_var(df=df,
                                    hue_var_name=hue_var_name,
                                    x_index_var_name=x_index_var_name,
                                    y_var_name=ylabel,
                                    add_zero_line=True,
                                    meanline=meanline,
                                    is_heatmap_colors=False,
                                    labels=prices.columns,
                                    legend_loc=legend_loc,
                                    x_rotation=x_rotation,
                                    yvar_format=yvar_format,
                                    ax=ax,
                                    **kwargs)
    return fig


def add_bnb_regime_shadows(ax: plt.Subplot,
                           data_df: pd.DataFrame = None,
                           benchmark: str = None,
                           pivot_prices: pd.Series = None,
                           regime_params: BenchmarkReturnsQuantileRegimeSpecs = None,
                           is_force_lim: bool = True,
                           alpha: float = 0.3,
                           **kwargs
                           ) -> None:

    if regime_params is None:
        regime_params = BenchmarkReturnsQuantileRegimeSpecs()
    regime_classifier = BenchmarkReturnsQuantilesRegime(regime_params=regime_params)

    if pivot_prices is not None:
        benchmark = pivot_prices.name
    elif benchmark is not None and data_df is not None:
        if benchmark in data_df.columns:
            pivot_prices = data_df[benchmark]
        else:
            raise KeyError(f"{benchmark} not in {data_df.columns}")
    else:
        raise ValueError(f"need pivot_prices or benchmark")

    regime_ids = regime_classifier.compute_sampled_returns_with_regime_id(prices=pivot_prices,
                                                                          benchmark=benchmark,
                                                                          **regime_params._asdict())

    regime_id_color = regime_classifier.class_data_to_colors(regime_data=regime_ids[RegimeClassifier.REGIME_COLUMN])

    # fill in the first date before the class date
    price_data_index = pivot_prices.index

    ax.axvspan(xmin=price_data_index[0],
               xmax=regime_ids.index[0],
               alpha=alpha,
               color=regime_id_color.loc[regime_id_color.index[0]], lw=0)
    for date_1, date in zip(regime_ids.index[:-1], regime_ids.index[1:]):
        ax.axvspan(xmin=date_1, xmax=date, alpha=alpha, color=regime_id_color[date], lw=0)
    if is_force_lim:
        ax.set_xlim(price_data_index[0], regime_ids.index[-1])


class UnitTests(Enum):
    BNB_REGIME = 1
    VOL_REGIME = 2
    BNB_REGIME_SHADOWS = 3
    BNB_PERF_TABLE = 4
    AVG_PLOT = 5


def run_unit_test(unit_test: UnitTests):

    from qis.test_data import load_etf_data
    prices = load_etf_data().dropna()

    kwargs = dict(var_format='{:.1f}')

    perf_params = PerfParams()

    if unit_test == UnitTests.BNB_REGIME:
        regime_params = BenchmarkReturnsQuantileRegimeSpecs()
        regime_classifier = BenchmarkReturnsQuantilesRegime(regime_params=regime_params)
        regime_ids = regime_classifier.compute_sampled_returns_with_regime_id(prices=prices, benchmark='SPY')
        print(f"regime_ids:\n{regime_ids}")

        cond_perf_table, regime_datas = regime_classifier.compute_regimes_pa_perf_table(prices=prices,
                                                                                        benchmark='SPY',
                                                                                        perf_params=perf_params)
        print(f"regime_means:\n{cond_perf_table}")
        print(f"regime_pa:\n{regime_datas}")

        fig, ax = plt.subplots(1, 1, figsize=(2.5, 2.5), tight_layout=True)

        plot_regime_data(regime_classifier=regime_classifier,
                         drop_sharpe_from_labels=True,
                         prices=prices,
                         benchmark='SPY',
                         perf_params=perf_params,
                         title='(A) Weekly roll',
                         is_use_vbar=True,
                         is_add_totals=False,
                         add_bar_values=False,
                         fontsize=8,
                         ncol=3,
                         bbox_to_anchor=(0.5, 1.12),
                         pad=15,
                         ax=ax,
                         **kwargs)

        plot_regime_data(regime_classifier=regime_classifier,
                         prices=prices,
                         benchmark='SPY',
                         perf_params=perf_params,
                         is_use_vbar=False,
                         bbox_to_anchor=None,
                         **kwargs)

    elif unit_test == UnitTests.VOL_REGIME:
        regime_params = VolQuantileRegimeSpecs()
        perf_params = PerfParams()
        regime_classifier = BenchmarkVolsQuantilesRegime(regime_params=regime_params)
        regime_ids = regime_classifier.compute_sampled_returns_with_regime_id(prices=prices, benchmark='SPY')
        print(f"regime_ids:\n{regime_ids}")

        cond_perf_table, regime_datas = regime_classifier.compute_regimes_pa_perf_table(prices=prices,
                                                                                   benchmark='SPY',
                                                                                   perf_params=perf_params)
        print(f"regime_means:\n{cond_perf_table}")
        print(f"regime_pa:\n{regime_datas}")

        plot_regime_data(regime_classifier=regime_classifier,
                         prices=prices,
                         benchmark='SPY',
                         perf_params=perf_params,
                         is_use_vbar=True,
                         **kwargs)

        plot_regime_data(regime_classifier=regime_classifier,
                         prices=prices,
                         benchmark='SPY',
                         perf_params=perf_params,
                         is_use_vbar=False,
                         **kwargs)

    elif unit_test == UnitTests.BNB_REGIME_SHADOWS:
        import qis.plots.time_series as pts
        with sns.axes_style('white'):
            fig, ax = plt.subplots(1, 1, figsize=(10, 10), constrained_layout=True)
            pts.plot_time_series(df=prices, ax=ax)

            add_bnb_regime_shadows(ax=ax,
                                   data_df=prices,
                                   benchmark='SPY',
                                   regime_params=BenchmarkReturnsQuantileRegimeSpecs(),
                                   perf_params=PerfParams())

    elif unit_test == UnitTests.BNB_PERF_TABLE:
        df = compute_bnb_regimes_pa_perf_table(prices=prices,
                                               benchmark='SPY',
                                               regime_params=BenchmarkReturnsQuantileRegimeSpecs(),
                                               perf_params=PerfParams())
        print(df)

    elif unit_test == UnitTests.AVG_PLOT:
        regime_params = VolQuantileRegimeSpecs()
        perf_params = PerfParams()
        regime_classifier = BenchmarkVolsQuantilesRegime(regime_params=regime_params)

        with sns.axes_style('white'):
            fig, ax = plt.subplots(1, 1, figsize=(7, 7), constrained_layout=True)
            plot_regime_boxplot(regime_classifier=regime_classifier,
                                prices=prices,
                                benchmark='SPY',
                                perf_params=perf_params,
                                ax=ax,
                                **kwargs)

    plt.show()


if __name__ == '__main__':

    unit_test = UnitTests.BNB_REGIME

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)



# biult in
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
from enum import Enum

# qis
import qis.file_utils as fu
import qis.utils.dates as da
import qis.models.linear.corr_cov_matrix as cor
import qis.plots.time_series as pts

# data
import yfinance as yf


def get_compute_corrs(prices: pd.DataFrame,
                      time_period: da.TimePeriod = None,
                      span: int = 33
                      ) -> pd.DataFrame:
    returns = prices.pct_change()
    corrs = cor.compute_ewm_corr_df(df=returns,
                                    corr_matrix_output=cor.CorrMatrixOutput.TOP_ROW,
                                    ewm_lambda=1.0-2.0/(span+1.0))

    if time_period is not None:
        corrs = time_period.locate(corrs)

    return corrs


class UnitTests(Enum):
    CORRS_SPY = 1
    CORRS_ALL = 2


def run_unit_test(unit_test: UnitTests):
    tickers = ['SPY', 'BTC-USD']
    prices = yf.download(tickers, start=None, end=None)['Adj Close'][tickers].dropna()
    prices = prices.rename({'BTC-USD': 'BTC'}, axis=1)
    time_period = da.TimePeriod('01Jan2021', None)

    if unit_test == UnitTests.CORRS_SPY:
        corrs = get_compute_corrs(prices=prices, span=33, time_period=time_period)
        trend = corrs.rolling(21).mean()
        df = pd.concat([corrs.iloc[:, 0].rename('Correlation'), trend.iloc[:, 0].rename('1m rolling average')], axis=1)
        fig = pts.plot_time_series(df=df,
                                   trend_line=pts.TrendLine.AVERAGE,
                                   legend_stats=pts.LegendStats.AVG_LAST,
                                   var_format='{:.2%}')

    elif unit_test == UnitTests.CORRS_ALL:

        returns = prices.pct_change()
        # returns = np.log(prices).diff()
        corr_1m = returns['BTC'].rolling(21).corr(returns['SPY']).rename('1m')
        corr_3m = returns['BTC'].rolling(63).corr(returns['SPY']).rename('3m')
        corr_1y = returns['BTC'].rolling(252).corr(returns['SPY']).rename('1y')
        corr_3y = returns['BTC'].rolling(3*252).corr(returns['SPY']).rename('3y')

        corrs = pd.concat([corr_1m, corr_3m, corr_1y, corr_3y], axis=1).dropna()
        corrs = time_period.locate(corrs)
        fu.save_df_to_excel(data=corrs, file_name='btc_spy_corr')

        with sns.axes_style("darkgrid"):
            fig, ax = plt.subplots(1, 1, figsize=(15, 8), tight_layout=True)

            pts.plot_time_series(df=corrs,
                                 trend_line=pts.TrendLine.AVERAGE,
                                 legend_stats=pts.LegendStats.AVG_LAST_SCORE,
                                 var_format='{:.0%}',
                                 fontsize=14,
                                 title='Rolling correlation of daily returns between BTC-S&P500 as function of rolling window',
                                 ax=ax)
        fu.save_fig(fig=fig, file_name='btc_all_corr')

    plt.show()


if __name__ == '__main__':

    unit_test = UnitTests.CORRS_ALL

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)
        


"""
different methods to generate portfolio weights from df
"""
import numpy as np
import pandas as pd
from typing import Union, Optional, Dict
from enum import Enum

# qis
import qis.utils.df_groups as dfg


class WeightMethod(str, Enum):
    EQUAL_WEIGHT = 'EqualWight'
    PROPORTIONAL = 'Proportional'
    SQRT_PROPORTIONAL = 'Sqrt '


def compute_long_only_portfolio_weights(data: pd.DataFrame,
                                        weight_method: WeightMethod = WeightMethod.EQUAL_WEIGHT
                                        ) -> pd.DataFrame:
    """
    compute [0, 1] weights data
    """
    if weight_method == WeightMethod.EQUAL_WEIGHT:
        weights = df_to_equal_weight_allocation(df=data)
    elif weight_method == WeightMethod.PROPORTIONAL:
        weights = df_to_weight_allocation_sum1(df=data)
    elif weight_method == WeightMethod.SQRT_PROPORTIONAL:
        weights = df_to_weight_allocation_sum1(df=np.sqrt(data))
    else:
        raise TypeError(f"not implemented method {weight_method}")
    return weights


def df_to_equal_weight_allocation(df: Union[pd.Series, pd.DataFrame],
                                  freq: str = None,
                                  index: pd.DatetimeIndex = None
                                  ) -> Union[pd.Series, pd.DataFrame]:
    """
    equal weight accounting for nans in df
    """
    equal_weight_allocation = df_to_weight_allocation_sum1(df=df_nans_to_one_zero(df=df))
    if freq is not None:
        equal_weight_allocation = equal_weight_allocation.asfreq(freq, method='ffill')
    elif index is not None:
        equal_weight_allocation = equal_weight_allocation.reindex(index, method='ffill')
    return equal_weight_allocation


def df_to_weight_allocation_sum1(df: Union[pd.Series, pd.DataFrame]) -> Union[pd.Series, pd.DataFrame]:
    """
    normalized rows by cross-sectional sum
    """
    if isinstance(df, pd.Series):
        weights = df.divide(np.nansum(df, axis=0)).fillna(0.0)
    else:
        weights = df.divide(np.nansum(df, axis=1, keepdims=True)).fillna(0.0)
    return weights


def df_to_long_only_allocation_sum1(df: Union[pd.Series, pd.DataFrame]) -> Union[pd.Series, pd.DataFrame]:
    """
    normalized rows by cross-sectional sum
    """
    weighted_score = np.where(df.to_numpy() > 0.0, df, 0.0)
    if isinstance(df, pd.Series):
        weights = weighted_score / np.nansum(weighted_score)
        weights = pd.Series(weights, index=df.index).fillna(0.0)
    else:
        weights = weighted_score / np.nansum(weighted_score, axis=1, keepdims=True)
        weights = pd.DataFrame(weights, index=df.index, columns=df.columns).fillna(0.0)
    return weights


def df_nans_to_one_zero(df: pd.DataFrame) -> pd.DataFrame:
    """
    return 1 if is set is not None and 0 otherwise
    """
    data = np.where(np.isfinite(df.to_numpy(dtype=np.float64)), 1.0, 0.0)
    return pd.DataFrame(data=data, index=df.index, columns=df.columns)


def mult_df_columns_with_vector(df: pd.DataFrame,
                                vector: pd.Series,
                                is_norm: bool = False,
                                nan_fill_zero: bool = False
                                ) -> pd.DataFrame:
    """
    multiply data set with vector column-wise and normalize accounting for nans in data
    data can be indicators data with False/True
    """
    conv = df.multiply(vector)

    if is_norm:  # nan sum of row of all nans is zero
        nump_data = conv.to_numpy(dtype=np.float64)
        column_sum = np.nansum(nump_data, axis=1, keepdims=True)  # column vector: column_sum=0.0 if all rows are nans
        nan_ind = np.all(np.isnan(nump_data), axis=1, keepdims=True)  # column vector to trace rows with all nans
        div_cond = np.logical_and(np.isclose(column_sum, 0.0) == False, nan_ind == False)
        conv = np.divide(conv, column_sum, where=div_cond)  # divide by column sum where possible
        if np.any(nan_ind):
            conv.loc[nan_ind.flatten(), :] = np.nan  # rows with nans = nans

    if nan_fill_zero:
        conv = conv.fillna(0)

    return conv


def mult_df_columns_with_vector_group(df: pd.DataFrame,
                                      vector: pd.Series,
                                      group_data: pd.Series,
                                      is_norm: bool = False,
                                      nan_fill_zero: bool = False,
                                      return_df: bool = False
                                      ) -> Union[Dict[str, pd.DataFrame], pd.DataFrame]:
    """
    convolve/multiply data set with vector column-wise and normalize with grouping accounting for nans in data
    """
    group_dict = dfg.split_df_by_groups(df=df, group_data=group_data)
    group_conv = {}
    for group, g_data in group_dict.items():
        group_conv[group] = mult_df_columns_with_vector(df=g_data,
                                                        vector=vector.loc[g_data.columns],
                                                        is_norm=is_norm,
                                                        nan_fill_zero=nan_fill_zero)
    if return_df:
        group_conv = pd.concat([v for k, v in group_conv.items()], axis=1)[df.columns]

    return group_conv


def df_to_top_n_indicators(df: Union[pd.Series, pd.DataFrame],
                           num_top_assets: int = 15
                           ) -> Union[pd.Series, pd.DataFrame]:
    """
    assign unit weight to ranked rows
    """
    def series_to_top_n_indicators(data: pd.Series) -> pd.Series:
        ranked_row = data.sort_values(ascending=False)
        ranked_row[:num_top_assets], ranked_row[num_top_assets:] = 1.0, 0.0
        ranked_row = ranked_row.sort_index()
        return ranked_row

    if isinstance(df, pd.Series):
        ranked_data = series_to_top_n_indicators(data=df)[df.index]
    else:
        columns = df.columns.copy()
        ranked_rows = {}
        for idx, row in df.iterrows():
            ranked_rows[idx] = series_to_top_n_indicators(data=row)
        ranked_data = pd.DataFrame.from_dict(ranked_rows, orient='index')[columns]
    return ranked_data


def fill_long_short_signal(rank_data: np.ndarray,
                           leg_size: int
                           ) -> np.ndarray:
    """
    place +1 for top, -1 for bottom
    """
    signal0 = np.zeros_like(rank_data)
    signal1 = np.ones_like(rank_data)
    bottom_quantile = leg_size
    upper_quantile = rank_data.shape[0] - leg_size - 1
    signal = np.where(rank_data > upper_quantile, signal1, signal0)  # +1 top ranks
    signal = np.where(rank_data < bottom_quantile, -signal1, signal)  # -1 smallest quantile
    return signal


def compute_long_short_ind_by_row(data: pd.DataFrame,
                                  cut_off_quantile: float = 0.1,
                                  min_leg_size: int = 2,
                                  leg_size: Optional[int] = None
                                  ) -> pd.DataFrame:
    """
    get cross sectional indicators
    """
    if len(data.columns) == 1:
        raise ValueError(f"one column is not supported for ranking")
    elif len(data.columns) == 2:
        leg_size = 0
    else:
        pass

    # compute quantiles
    if leg_size is None:
        leg_size = np.maximum(np.floor(cut_off_quantile * len(data.columns)),
                              min_leg_size)

    rank_data = data.rank(axis=1, method='first', na_option='keep', ascending=False)

    signal = fill_long_short_signal(rank_data=rank_data.to_numpy(), leg_size=leg_size)
    cross_sectional_indicators = pd.DataFrame(data=signal, columns=data.columns, index=data.index)

    return cross_sectional_indicators


def compute_long_short_ind(data: np.ndarray,
                           cut_off_quantile: float = 0.1,
                           min_leg_size: int = 2,
                           leg_size: Optional[int] = None
                           ) -> np.ndarray:
    """
    row wise operator to assing -/+1 for ranked data in df
    """
    if not data.ndim == 1:
        raise ValueError(f"ndim must be 1")

    n = data.shape[0]
    # compute quantiles
    if leg_size is None:
        leg_size = np.maximum(np.floor(cut_off_quantile * n), min_leg_size)

    rank_data = pd.DataFrame(data).rank(axis=1, method='first', na_option='keep', ascending=False)

    signal = fill_long_short_signal(rank_data=rank_data.to_numpy(), leg_size=leg_size)

    return signal


class UnitTests(Enum):
    CONV = 1
    EW_ALLOC = 2
    AC_EQUAL_WEIGHT_ALLOCATION = 3


def run_unit_test(unit_test: UnitTests):

    import qis.utils.df_ops as dfo

    if unit_test == UnitTests.CONV:
        constituent_dict = {'f1': np.array((0, np.nan, 1, 1)),
                            'f2': np.array((0, np.nan, np.nan, 1)),
                            'f3': np.array((0, np.nan, 1, 1)),
                            'f4': np.array((0, np.nan, 1, 1)),
                            'f5': np.array((0, np.nan, np.nan, 1)),
                            'f6': np.array((0, np.nan, 1, 1))}

        constituent_prices = pd.DataFrame.from_dict(constituent_dict, orient='columns')
        constituent_data = dfo.df_indicator_like(constituent_prices)
        print(constituent_data)

        desc_dict = {'f1': ('eq', 1.0),
                     'f2': ('fi', 0.5),
                     'f3': ('fi', 1.5),
                     'f4': ('fx', 1.0),
                     'f5': ('fx', 2.0),
                     'f6': ('fx', 3.0)}

        desc_data = pd.DataFrame.from_dict(desc_dict, orient='index', columns=['ac', 'acw'])
        print(desc_data)

        conv = mult_df_columns_with_vector(df=constituent_data, vector=desc_data['acw'])
        print(f"conv:\n{conv}")

        conv_norm = mult_df_columns_with_vector(df=constituent_data, vector=desc_data['acw'], is_norm=True)
        print(f"conv_norm:\n{conv_norm}")

        ac_conv = mult_df_columns_with_vector_group(df=constituent_data, vector=desc_data['acw'],
                                                    group_data=desc_data['ac'])
        for ac, data in ac_conv.items():
            print(f"{ac}:\n{data}")

        ac_conv = mult_df_columns_with_vector_group(df=constituent_data, vector=desc_data['acw'],
                                                    group_data=desc_data['ac'],
                                                    is_norm=True)
        for ac, data in ac_conv.items():
            print(f"norm {ac}:\n{data}")

        ac_conv_pd = mult_df_columns_with_vector_group(df=constituent_data, vector=desc_data['acw'],
                                                       group_data=desc_data['ac'],
                                                       is_norm=True,
                                                       return_df=True)
        print(f"ac_conv_pd:\n{ac_conv_pd}")

    elif unit_test == UnitTests.EW_ALLOC:
        constituent_dict = {'f1': np.array((0, np.nan, 2, 3)),
                            'f2': np.array((0, np.nan, np.nan, 4)),
                            'f3': np.array((0, np.nan, 1, 1)),
                            'f4': np.array((0, np.nan, 1, 1)),
                            'f5': np.array((0, np.nan, np.nan, 1)),
                            'f6': np.array((0, np.nan, 1, 1))}

        constituent_prices = pd.DataFrame.from_dict(constituent_dict, orient='columns')
        print(constituent_prices)
        weights = df_to_equal_weight_allocation(constituent_prices)
        print(weights)

    elif unit_test == UnitTests.AC_EQUAL_WEIGHT_ALLOCATION:
        universe_data = dict(SPY='Equities',
                             QQQ='Equities',
                             EEM='Equities',
                             TLT='Bonds',
                             IEF='Bonds',
                             LQD='Credit',
                             HYG='Credit',
                             GLD='Gold')
        # each asset class has equal weight allocation
        group_data = pd.Series(universe_data)
        constituents = pd.DataFrame(1.0, columns=group_data.index, index=['date1', 'date2'])
        constituents.loc['date1', 'EEM'] = np.nan
        constituents.loc['date2', 'IEF'] = np.nan
        within_ac_weight = pd.Series(1.0, index=group_data.index)
        within_ac_weight['HYG'] = 0.5
        rel_ac_weight_by_ints = mult_df_columns_with_vector_group(df=constituents,
                                                                  vector=within_ac_weight,
                                                                  group_data=group_data,
                                                                  is_norm=True, nan_fill_zero=True,
                                                                  return_df=True)
        rel_ac_weight_by_ints = df_to_weight_allocation_sum1(df=rel_ac_weight_by_ints)
        print(rel_ac_weight_by_ints)
        print(rel_ac_weight_by_ints.sum(1))


if __name__ == '__main__':

    unit_test = UnitTests.AC_EQUAL_WEIGHT_ALLOCATION

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)



"""
to use pybloqs for pandas > 2.x
locate file "...\Lib\site-packages\pybloqs\jinja\table.html"
change line 44 from:
{% for col_name, cell in row.iteritems() %}
to:
{% for col_name, cell in row.items() %}
"""

# packages
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
from enum import Enum
import pybloqs as p

# qis
import yfinance as yf
import qis
from qis import PerfStat, local_path


# define entries to show in ra perf table
RA_TABLE_COLUMNS = [#PerfStat.START_DATE,
    #PerfStat.END_DATE,
    #PerfStat.START_PRICE,
    #PerfStat.END_PRICE,
    PerfStat.TOTAL_RETURN,
    PerfStat.PA_RETURN,
    PerfStat.VOL,
    PerfStat.SHARPE_RF0,
    PerfStat.SHARPE_EXCESS,
    PerfStat.MAX_DD,
    PerfStat.MAX_DD_VOL,
    PerfStat.SKEWNESS,
    PerfStat.KURTOSIS]


def generate_performances(prices: pd.DataFrame,
                          regime_benchmark_str: str,
                          perf_params: qis.PerfParams = None,
                          heatmap_freq: str = 'YE',
                          **kwargs
                          ) -> None:

    local_kwargs = dict(digits_to_show=1, framealpha=0.75, perf_stats_labels=qis.PerfStatsLabels.DETAILED_WITH_DD.value)
    kwargs = qis.update_kwargs(kwargs, local_kwargs)

    figsize = qis.get_df_table_size(df=pd.DataFrame(),
                                    min_cols=len(RA_TABLE_COLUMNS),
                                    min_rows=len(prices.columns),
                                    scale_rows=0.5,
                                    scale_cols=0.5)
    fig1, ax = plt.subplots(1, 1, figsize=(14, 1.1), tight_layout=True)

    qis.plot_ra_perf_table(prices=prices,
                           perf_columns=RA_TABLE_COLUMNS,
                           perf_params=perf_params,
                           # title=f"Risk-adjusted performance: {qis.get_time_period_label(prices, date_separator='-')}",
                           fontsize=8,
                           ax=ax,
                           **kwargs)

    ra_table = qis.get_ra_perf_columns(prices=prices,
                                       perf_columns=RA_TABLE_COLUMNS,
                                       perf_params=perf_params,
                                       drop_index=True,
                                       **kwargs)

    fig2, ax = plt.subplots(1, 1, figsize=(7, qis.calc_table_height(num_rows=len(prices.columns)+5, scale=0.5)), tight_layout=True)
    qis.plot_periodic_returns_table(prices=prices,
                                    freq=heatmap_freq,
                                    ax=ax,
                                    title=f"Periodic performance: {qis.get_time_period_label(prices, date_separator='-')}",
                                    total_name='Total',
                                    **qis.update_kwargs(kwargs, dict(square=False, x_rotation=90)))

    with sns.axes_style("darkgrid"):
        fig3, ax = plt.subplots(1, 1, figsize=(8, 6))
        qis.plot_prices(prices=prices,
                        regime_benchmark_str=regime_benchmark_str,
                        perf_params=perf_params,
                        title=f"Time series of $1 invested: {qis.get_time_period_label(prices, date_separator='-')}",
                        ax=ax,
                        **kwargs)

        fig4, axs = plt.subplots(2, 1, figsize=(7, 7))
        qis.plot_prices_with_dd(prices=prices,
                                regime_benchmark_str=regime_benchmark_str,
                                perf_params=perf_params,
                                title=f"Time series of $1 invested: {qis.get_time_period_label(prices, date_separator='-')}",
                                axs=axs,
                                **kwargs)

        fig5, ax = plt.subplots(1, 1, figsize=(8, 6))
        qis.plot_scatter_regression(prices=prices,
                                    regime_benchmark_str=regime_benchmark_str,
                                    regime_params=qis.BenchmarkReturnsQuantileRegimeSpecs(freq=perf_params.freq_reg),
                                    perf_params=perf_params,
                                    title=f"Regime Conditional Regression: {qis.get_time_period_label(prices, date_separator='-')}",
                                    ax=ax,
                                    **kwargs)

    bkwargs_title = {'title_wrap': True, 'text-align': 'left', 'color': 'blue', 'font_size': "12px",
                     'title_level': 3, 'line-height': 0.1, 'inherit_cfg': False}
    bkwargs_text = {'title_wrap': True, 'text-align': 'left', 'font_size': "10px",
                    'title_level': 3, 'line-height': 0.1}
    bkwargs_fig = {'title_wrap': True, 'text-align': 'left', 'font_size': "10px",
                   'title_level': 3, 'line-height': 0.1}
    b_fig_ra = p.Block(
        [p.Paragraph(f"Block with Risk-adjusted performance", **bkwargs_title),
         p.Paragraph("Description", **bkwargs_text),
         p.Block(fig1, **bkwargs_fig)],
        **bkwargs_text)

    r = p.VStack([p.Block("Strategy description", title="Backtest strategy"),
                  b_fig_ra,
                  p.Block(fig2, title='New'),
                  p.Block(ra_table),
                  p.Block(fig3), p.Block(fig4), p.Block(fig5)])

    report_file_name = f"{local_path.get_output_path()}PyBloqs_Factsheet_{pd.Timestamp.now().strftime('%Y%m%d_%H%M')}.pdf"
    print(f"saved to {report_file_name}")
    r.save(report_file_name)


class UnitTests(Enum):
    ETF_DATA = 1
    CRYPTO_DATA = 2


def run_unit_test(unit_test: UnitTests):

    ust_3m_rate = yf.download('^IRX', start=None, end=None)['Adj Close'].dropna() / 100.0

    if unit_test == UnitTests.ETF_DATA:
        regime_benchmark_str = 'SPY'
        tickers = [regime_benchmark_str, 'QQQ', 'EEM', 'TLT', 'IEF', 'LQD', 'HYG', 'SHY', 'GLD']

    elif unit_test == UnitTests.CRYPTO_DATA:
        regime_benchmark_str = 'BTC-USD'
        tickers = [regime_benchmark_str, 'ETH-USD', 'SOL-USD']

    else:
        raise NotImplementedError

    is_long_period = False
    if is_long_period:
        time_period = None
        perf_params = qis.PerfParams(freq='W-WED', freq_reg='QE', freq_drawdown='B', rates_data=ust_3m_rate)
        kwargs = dict(x_date_freq='YE', heatmap_freq='YE', date_format='%Y', perf_params=perf_params)
    else:
        time_period = qis.TimePeriod('31Dec2021', '23Jan2023')
        perf_params = qis.PerfParams(freq='W-WED', freq_reg='W-WED', freq_drawdown='B', rates_data=ust_3m_rate)
        kwargs = dict(x_date_freq='ME', heatmap_freq='ME', date_format='%b-%y', perf_params=perf_params)

    prices = yf.download(tickers, start=None, end=None)['Adj Close'].dropna()

    if time_period is not None:
        prices = time_period.locate(prices)

    generate_performances(prices=prices,
                          regime_benchmark_str=regime_benchmark_str,
                          **kwargs)

    plt.show()


if __name__ == '__main__':

    unit_test = UnitTests.ETF_DATA

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)




from qis.utils.dates import (
    TimePeriod,
    generate_dates_schedule,
    generate_rebalancing_indicators,
    set_rebalancing_timeindex_on_given_timeindex,
    generate_sample_dates,
    get_month_days,
    get_period_days,
    get_sample_dates_idx,
    get_time_period,
    get_time_period_label,
    get_time_period_shifted_by_years,
    get_current_time_with_tz,
    get_weekday,
    get_year_quarter,
    get_ytd_time_period,
    get_time_to_maturity,
    infer_an_from_data,
    is_leap_year,
    months_between,
    separate_number_from_string,
    shift_date_by_day,
    shift_dates_by_n_years,
    shift_dates_by_year,
    shift_time_period_by_days,
    split_df_by_freq,
    generate_fixed_maturity_rolls,
    min_timestamp,
    truncate_prior_to_start
)


from qis.utils.df_agg import (
    abssum,
    abssum_negative,
    abssum_positive,
    agg_data_by_axis,
    agg_dfs,
    agg_median_mad,
    get_signed_np_data,
    nanmean,
    nanmean_clip,
    nanmean_positive,
    nanmedian,
    nansum,
    nansum_clip,
    nansum_negative,
    nansum_positive,
    sum_weighted,
    last_row,
    compute_df_desc_data
)

from qis.utils.df_cut import (
    add_classification,
    add_hue_fixed_years,
    add_hue_years,
    add_quantile_classification,
    sort_index_by_hue,
    x_bins_cut
)

from qis.utils.df_freq import (
    agg_remained_data_on_right,
    df_asfreq,
    df_index_to_str,
    df_resample_at_freq,
    df_resample_at_int_index,
    df_resample_at_other_index
)

from qis.utils.df_groups import (
    agg_df_by_group_with_avg,
    agg_df_by_groups,
    agg_df_by_groups_ax1,
    fill_df_with_group_avg,
    get_group_dict,
    sort_df_by_index_group,
    split_df_by_groups,
    set_group_loadings
)

from qis.utils.df_melt import (
    melt_df_by_columns,
    melt_paired_df,
    melt_scatter_data_with_xdata,
    melt_scatter_data_with_xvar,
    melt_signed_paired_df
)

from qis.utils.df_ops import (
    align_df1_to_df2,
    align_dfs_dict_with_df,
    compute_last_score,
    df12_merge_with_tz,
    df_indicator_like,
    df_joint_indicator,
    df_ones_like,
    df_time_dict_to_pd,
    df_zero_like,
    dfs_indicators,
    dfs_to_upper_lower_diag,
    drop_first_nan_data,
    factor_dict_to_asset_dict,
    get_first_before_nonnan_index,
    get_first_last_nonnan_index,
    get_first_nonnan_values,
    get_last_nonnan_values,
    get_last_nonnan,
    merge_on_column,
    compute_nans_zeros_ratio_after_first_non_nan,
    reindex_upto_last_nonnan,
    multiply_df_by_dt,
    norm_df_by_ax_mean,
    np_txy_tensor_to_pd_dict
)

from qis.utils.df_str import (
    date_to_str,
    df_all_to_str,
    df_index_to_str,
    df_to_numeric,
    df_to_str,
    df_with_ci_to_str,
    float_to_str,
    get_fmt_str,
    join_str_series,
    series_to_date_str,
    series_to_numeric,
    series_to_str,
    series_values_to_str,
    str_to_float,
    timeseries_df_to_str,
    idx_to_alphabet
)

from qis.utils.df_to_weights import (
    compute_long_short_ind,
    compute_long_short_ind_by_row,
    df_nans_to_one_zero,
    df_to_equal_weight_allocation,
    df_to_top_n_indicators,
    df_to_weight_allocation_sum1,
    df_to_long_only_allocation_sum1,
    fill_long_short_signal,
    compute_long_only_portfolio_weights,
    mult_df_columns_with_vector,
    mult_df_columns_with_vector_group
)
from qis.utils.df_to_scores import (
    df_to_max_score,
    df_to_cross_sectional_score)

from qis.utils.generic import (
    ValueType,
    ColVar,
    ColumnData,
    column_datas_to_df,
    EnumMap,
    DotDict
)


from qis.utils.np_ops import (
    compute_expanding_power,
    compute_histogram_data,
    compute_paired_signs,
    covar_to_corr,
    find_nearest,
    to_nearest_values,
    np_array_to_df_columns,
    np_array_to_df_index,
    np_array_to_matrix,
    np_array_to_n_column_array,
    np_array_to_t_rows_array,
    np_cumsum,
    np_nanstd,
    np_get_sorted_idx,
    np_matrix_add_array,
    np_nonan_weighted_avg,
    np_shift,
    running_mean,
    to_finite_np,
    to_finite_ratio,
    to_finite_reciprocal,
    repeat_by_columns,
    repeat_by_rows
)

from qis.utils.ols import (
    estimate_alpha_beta_paired_dfs,
    estimate_ols_alpha_beta,
    fit_ols,
    get_ols_x,
    reg_model_params_to_str
)

from qis.utils.sampling import (
    TrainLivePeriod,
    TrainLiveSamples,
    get_data_samples_df,
    split_to_samples,
    split_to_train_live_samples
)

from qis.utils.struct_ops import (
    assert_list_subset,
    assert_list_unique,
    flatten,
    flatten_dict_tuples,
    list_diff,
    list_intersection,
    list_to_unique_and_dub,
    merge_lists_unique,
    move_item_to_first,
    separate_number_from_string,
    split_dict,
    to_flat_list,
    update_kwargs
)

from qis.file_utils import timer





"""
generate portfolio factsheet report using PortfolioData object
PortfolioData object can be generated by a backtester or actucal strategy data
with comparison to 1-2 cash benchmarks
PortfolioData can contain either simulated or actual portfolio data
"""
# packages
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
from typing import Tuple, Optional, List
import qis as qis
from qis import TimePeriod, PerfParams, BenchmarkReturnsQuantileRegimeSpecs
from qis.portfolio.portfolio_data import PortfolioData
from qis.portfolio.reports.config import PERF_PARAMS, REGIME_PARAMS


def generate_strategy_factsheet(portfolio_data: PortfolioData,
                                benchmark_prices: pd.DataFrame,
                                time_period: TimePeriod,
                                perf_params: PerfParams = PERF_PARAMS,
                                regime_params: BenchmarkReturnsQuantileRegimeSpecs = REGIME_PARAMS,
                                regime_benchmark: str = None,  # default is set to benchmark_prices.columns[0]
                                exposures_freq: Optional[str] = 'W-WED',  #'W-WED',
                                turnover_rolling_period: int = 260,
                                turnover_freq: str = 'B',
                                turnover_title: Optional[str] = None,
                                factor_beta_span: int = 52,
                                beta_freq: str = 'W-WED',
                                factor_beta_title: Optional[str] = None,
                                factor_attribution_title: Optional[str] = None,
                                add_benchmarks_to_navs: bool = False,
                                figsize: Tuple[float, float] = (8.3, 11.7),  # A4 for portrait
                                fontsize: int = 4,
                                add_current_position_var_risk_sheet: bool = True,
                                add_grouped_exposures: bool = False,
                                add_grouped_cum_pnl: bool = False,
                                is_1y_exposures: bool = False,
                                is_grouped: Optional[bool] = None,
                                **kwargs
                                ) -> List[plt.Figure]:
    # align
    benchmark_prices = benchmark_prices.reindex(index=portfolio_data.nav.index, method='ffill')
    if regime_benchmark is None:
        regime_benchmark = benchmark_prices.columns[0]

    if is_grouped is None:
        if len(portfolio_data.group_data.unique()) < 10:  # otherwise tables look too bad
            is_grouped = True
        else:
            is_grouped = False

    fig = plt.figure(figsize=figsize, constrained_layout=True)
    gs = fig.add_gridspec(nrows=14, ncols=4, wspace=0.0, hspace=0.0)

    plot_kwargs = dict(fontsize=fontsize,
                       linewidth=0.5,
                       digits_to_show=1, sharpe_digits=2,
                       weight='normal',
                       markersize=1,
                       framealpha=0.75)
    kwargs = qis.update_kwargs(kwargs, plot_kwargs)
    fig.suptitle(f"{portfolio_data.nav.name} factsheet", fontweight="bold", fontsize=8, color='blue')

    # prices
    portfolio_nav = portfolio_data.get_portfolio_nav(time_period=time_period)
    if add_benchmarks_to_navs:
        benchmark_prices_ = benchmark_prices
    else:
        benchmark_prices_ = benchmark_prices[regime_benchmark]
    joint_prices = pd.concat([portfolio_nav, benchmark_prices_], axis=1).dropna()
    pivot_prices = joint_prices[regime_benchmark]
    ax = fig.add_subplot(gs[0:2, :2])
    qis.plot_prices(prices=joint_prices,
                    perf_params=perf_params,
                    title=f"Cumulative performance with background colors using bear/normal/bull regimes of {regime_benchmark} {regime_params.freq}-returns",
                    ax=ax,
                    **kwargs)
    qis.add_bnb_regime_shadows(ax=ax, pivot_prices=pivot_prices, regime_params=regime_params)
    qis.set_spines(ax=ax, bottom_spine=False, left_spine=False)

    # dd
    ax = fig.add_subplot(gs[2:4, :2])
    qis.plot_rolling_drawdowns(prices=joint_prices,
                               title='Running Drawdowns',
                               dd_legend_type=qis.DdLegendType.DETAILED,
                               ax=ax, **kwargs)
    qis.add_bnb_regime_shadows(ax=ax, pivot_prices=pivot_prices, regime_params=regime_params)
    qis.set_spines(ax=ax, bottom_spine=False, left_spine=False)

    # under watre
    ax = fig.add_subplot(gs[4:6, :2])
    qis.plot_rolling_time_under_water(prices=joint_prices,
                                      title='Running Time under Water',
                                      dd_legend_type=qis.DdLegendType.DETAILED,
                                      ax=ax, **kwargs)
    qis.add_bnb_regime_shadows(ax=ax, pivot_prices=pivot_prices, regime_params=regime_params)
    qis.set_spines(ax=ax, bottom_spine=False, left_spine=False)

    # exposures
    exposures = portfolio_data.get_weights(is_grouped=is_grouped, time_period=time_period, add_total=False)
    ax = fig.add_subplot(gs[6:8, :2])
    if exposures_freq is not None:
        exposures = exposures.resample(exposures_freq).last()
    qis.plot_stack(df=exposures,
                   use_bar_plot=True,
                   title='Exposures',
                   legend_stats=qis.LegendStats.AVG_NONNAN_LAST,
                   var_format='{:.1%}',
                   ax=ax,
                   **qis.update_kwargs(kwargs, dict(bbox_to_anchor=(0.5, 1.05), ncol=2)))
    qis.set_spines(ax=ax, bottom_spine=False, left_spine=False)

    # turnover
    ax = fig.add_subplot(gs[8:10, :2])

    turnover = portfolio_data.get_turnover(time_period=time_period, roll_period=turnover_rolling_period,
                                           freq=turnover_freq,
                                           is_grouped=is_grouped)
    freq = pd.infer_freq(turnover.index)
    turnover_title = turnover_title or f"{turnover_rolling_period}-period rolling {freq}-freq Turnover"
    qis.plot_time_series(df=turnover,
                         var_format='{:,.2%}',
                         # y_limits=(0.0, None),
                         legend_stats=qis.LegendStats.AVG_NONNAN_LAST,
                         title=turnover_title,
                         ax=ax,
                         **kwargs)
    qis.add_bnb_regime_shadows(ax=ax, pivot_prices=pivot_prices, regime_params=regime_params)
    qis.set_spines(ax=ax, bottom_spine=False, left_spine=False)

    # costs
    ax = fig.add_subplot(gs[10:12, :2])
    costs = portfolio_data.get_costs(time_period=time_period, roll_period=turnover_rolling_period)
    freq = pd.infer_freq(costs.index)
    costs_title = f"{turnover_rolling_period}-period rolling {freq}-freq Costs"
    qis.plot_time_series(df=costs,
                         var_format='{:,.2%}',
                         # y_limits=(0.0, None),
                         legend_stats=qis.LegendStats.AVG_NONNAN_LAST,
                         title=costs_title,
                         ax=ax,
                         **kwargs)
    qis.add_bnb_regime_shadows(ax=ax, pivot_prices=pivot_prices, regime_params=regime_params)
    qis.set_spines(ax=ax, bottom_spine=False, left_spine=False)

    # constituents
    ax = fig.add_subplot(gs[12:, :2])
    num_investable_instruments = portfolio_data.get_num_investable_instruments(time_period=time_period)
    qis.plot_time_series(df=num_investable_instruments,
                         var_format='{:,.0f}',
                         legend_stats=qis.LegendStats.FIRST_AVG_LAST,
                         title='Number of investable and invested instruments',
                         ax=ax,
                         **kwargs)
    qis.add_bnb_regime_shadows(ax=ax, pivot_prices=pivot_prices, regime_params=regime_params)
    qis.set_spines(ax=ax, bottom_spine=False, left_spine=False)

    # ra perf table
    if add_benchmarks_to_navs:
        benchmark_price1 = benchmark_prices
    else:
        benchmark_price1 = benchmark_prices[regime_benchmark]
    if is_grouped:
        ax = fig.add_subplot(gs[:2, 2:])
        portfolio_data.plot_ra_perf_table(ax=ax,
                                          benchmark_price=benchmark_price1,
                                          time_period=time_period,
                                          perf_params=perf_params,
                                          is_grouped=is_grouped,
                                          **qis.update_kwargs(kwargs, dict(fontsize=fontsize)))
    else:  # plot two tables
        ax = fig.add_subplot(gs[0, 2:])
        portfolio_data.plot_ra_perf_table(ax=ax,
                                          benchmark_price=benchmark_price1,
                                          time_period=time_period,
                                          perf_params=perf_params,
                                          is_grouped=is_grouped,
                                          **qis.update_kwargs(kwargs, dict(fontsize=fontsize)))
        ax = fig.add_subplot(gs[1, 2:])
        # change regression to weekly
        time_period1 = qis.get_time_period_shifted_by_years(time_period=time_period)
        if pd.infer_freq(benchmark_prices.index) in ['B', 'D']:
            local_kwargs = qis.update_kwargs(kwargs, dict(time_period=time_period1, alpha_an_factor=52, freq_reg='W-WED', fontsize=fontsize))
        else:
            local_kwargs = qis.update_kwargs(kwargs, dict(time_period=time_period1, fontsize=fontsize))
        portfolio_data.plot_ra_perf_table(ax=ax,
                                          benchmark_price=benchmark_price1,
                                          perf_params=perf_params,
                                          is_grouped=is_grouped,
                                          **local_kwargs)

    # heatmap
    ax = fig.add_subplot(gs[2:4, 2:])
    portfolio_data.plot_monthly_returns_heatmap(ax=ax,
                                                time_period=time_period,
                                                title='Monthly Returns',
                                                **qis.update_kwargs(kwargs, dict(fontsize=fontsize, date_format='%Y')))

    # periodic returns
    ax = fig.add_subplot(gs[4:6, 2:])
    local_kwargs = qis.update_kwargs(kwargs=kwargs,
                                     new_kwargs=dict(fontsize=fontsize, square=False, x_rotation=90, transpose=True))
    portfolio_data.plot_periodic_returns(benchmark_prices=benchmark_price1,
                                         is_grouped=is_grouped,
                                         time_period=time_period,
                                         ax=ax,
                                         **local_kwargs)

    # perf contributors
    with sns.axes_style("whitegrid"):
        ax = fig.add_subplot(gs[6:8, 2])
        portfolio_data.plot_contributors(ax=ax,
                                         time_period=time_period,
                                         title=f"Performance Contributors {time_period.to_str()}",
                                         **kwargs)
    with sns.axes_style("whitegrid"):
        ax = fig.add_subplot(gs[6:8, 3])
        time_period_1y = qis.get_time_period_shifted_by_years(time_period=time_period)
        portfolio_data.plot_contributors(ax=ax,
                                         time_period=time_period_1y,
                                         title=f"Performance Contributors {time_period_1y.to_str()}",
                                         **kwargs)

    # regime data
    portfolio_data.plot_regime_data(is_grouped=is_grouped,
                                    benchmark_price=benchmark_prices[regime_benchmark],
                                    time_period=time_period,
                                    perf_params=perf_params,
                                    regime_params=regime_params,
                                    ax=fig.add_subplot(gs[8:10, 2:]),
                                    **kwargs)

    # perf attribution
    local_kwargs = qis.update_kwargs(kwargs=kwargs, new_kwargs=dict(legend_loc=None))
    with sns.axes_style("whitegrid"):
        portfolio_data.plot_performance_attribution(time_period=time_period,
                                                    attribution_metric=qis.AttributionMetric.PNL,
                                                    ax=fig.add_subplot(gs[10:12, 2:]),
                                                    **local_kwargs)
        portfolio_data.plot_performance_attribution(time_period=time_period,
                                                    attribution_metric=qis.AttributionMetric.PNL_RISK,
                                                    ax=fig.add_subplot(gs[12:, 2:]),
                                                    **local_kwargs)

    figs = [fig]

    if add_current_position_var_risk_sheet:
        fig = plt.figure(figsize=figsize, constrained_layout=True)
        fig.suptitle(f"{portfolio_data.nav.name} current position, 95%-Var, and risk profile", fontweight="bold",
                     fontsize=8, color='blue')
        figs.append(fig)
        gs = fig.add_gridspec(nrows=14, ncols=4, wspace=0.0, hspace=0.0)

        # current position
        with sns.axes_style("whitegrid"):
            portfolio_data.plot_current_weights(is_grouped=False,
                                                ax=fig.add_subplot(gs[:2, :2]), **kwargs)
            portfolio_data.plot_current_weights(is_grouped=True,
                                                ax=fig.add_subplot(gs[:2, 2:]), **kwargs)

        # current var
        with sns.axes_style("whitegrid"):
            portfolio_data.plot_current_var(is_grouped=False, is_correlated=False,
                                            ax=fig.add_subplot(gs[2:4, :2]), **kwargs)
            portfolio_data.plot_current_var(is_grouped=True, is_correlated=False,
                                            ax=fig.add_subplot(gs[2:4, 2]), **kwargs)
            portfolio_data.plot_current_var(is_grouped=True, is_correlated=True,
                                            ax=fig.add_subplot(gs[2:4, 3]), **kwargs)

        # change in position
        with sns.axes_style("whitegrid"):
            portfolio_data.plot_last_weights_change(is_grouped=False,
                                                    ax=fig.add_subplot(gs[4:6, :2]), **kwargs)
            portfolio_data.plot_last_weights_change(is_grouped=True,
                                                    ax=fig.add_subplot(gs[4:6, 2:]), **kwargs)

        # benchmark betas
        ax = fig.add_subplot(gs[6:8, :2])
        factor_exposures = portfolio_data.compute_portfolio_benchmark_betas(benchmark_prices=benchmark_prices,
                                                                            time_period=time_period,
                                                                            beta_freq=beta_freq,
                                                                            factor_beta_span=factor_beta_span)
        factor_beta_title = factor_beta_title or f"Rolling {factor_beta_span}-period beta of {beta_freq}-freq returns"
        qis.plot_time_series(df=factor_exposures,
                             var_format='{:,.2f}',
                             legend_stats=qis.LegendStats.AVG_NONNAN_LAST,
                             title=factor_beta_title,
                             ax=ax,
                             **kwargs)
        qis.add_bnb_regime_shadows(ax=ax, pivot_prices=pivot_prices, regime_params=regime_params)
        qis.set_spines(ax=ax, bottom_spine=False, left_spine=False)

        # attribution
        ax = fig.add_subplot(gs[8:10, :2])
        factor_attribution = portfolio_data.compute_portfolio_benchmark_attribution(benchmark_prices=benchmark_prices,
                                                                                    beta_freq=beta_freq,
                                                                                    factor_beta_span=factor_beta_span,
                                                                                    time_period=time_period)
        factor_attribution_title = factor_attribution_title or f"Cumulative return attribution using rolling " \
                                                               f"{factor_beta_span}-period beta of {beta_freq}-freq returns"
        qis.plot_time_series(df=factor_attribution.cumsum(0),
                             var_format='{:,.0%}',
                             legend_stats=qis.LegendStats.LAST_NONNAN,
                             title=factor_attribution_title,
                             ax=ax,
                             **kwargs)
        qis.add_bnb_regime_shadows(ax=ax, pivot_prices=pivot_prices, regime_params=regime_params)
        qis.set_spines(ax=ax, bottom_spine=False, left_spine=False)

        # returns scatter
        with sns.axes_style("whitegrid"):
            portfolio_data.plot_returns_scatter(ax=fig.add_subplot(gs[6:8, 2:]),
                                                benchmark_price=benchmark_prices.iloc[:, 0],
                                                time_period=time_period,
                                                freq=perf_params.freq_reg,
                                                is_grouped=is_grouped,
                                                **kwargs)

        if len(benchmark_prices.columns) > 1:
            with sns.axes_style("whitegrid"):
                portfolio_data.plot_returns_scatter(ax=fig.add_subplot(gs[8:10, 2:]),
                                                    benchmark_price=benchmark_prices.iloc[:, 1],
                                                    time_period=time_period,
                                                    freq=perf_params.freq_reg,
                                                    is_grouped=is_grouped,
                                                    **kwargs)

        # var
        ax = fig.add_subplot(gs[10:12, :2])
        portfolio_data.plot_portfolio_grouped_var(ax=ax,
                                                  is_correlated=True,
                                                  time_period=time_period,
                                                  **kwargs)
        qis.add_bnb_regime_shadows(ax=ax, pivot_prices=pivot_prices, regime_params=regime_params)
        qis.set_spines(ax=ax, bottom_spine=False, left_spine=False)
        ax = fig.add_subplot(gs[10:12, 2:])
        portfolio_data.plot_portfolio_grouped_var(ax=ax,
                                                  is_correlated=False,
                                                  time_period=time_period,
                                                  **kwargs)
        qis.add_bnb_regime_shadows(ax=ax, pivot_prices=pivot_prices, regime_params=regime_params)
        qis.set_spines(ax=ax, bottom_spine=False, left_spine=False)

        # vol regime data
        ax = fig.add_subplot(gs[12:14, :2])
        portfolio_data.plot_vol_regimes(ax=ax,
                                        benchmark_price=benchmark_prices.iloc[:, 0],
                                        is_grouped=is_grouped,
                                        time_period=time_period,
                                        freq=regime_params.freq,
                                        regime_params=regime_params,
                                        **kwargs)
        if len(benchmark_prices.columns) > 1:
            ax = fig.add_subplot(gs[12:14, 2:])
            portfolio_data.plot_vol_regimes(ax=ax,
                                            benchmark_price=benchmark_prices.iloc[:, 1],
                                            is_grouped=is_grouped,
                                            time_period=time_period,
                                            freq=regime_params.freq,
                                            regime_params=regime_params,
                                            **kwargs)

    # set 1y time period for exposures
    if is_1y_exposures:
        time_period1 = qis.get_time_period_shifted_by_years(time_period=time_period)
        regime_params1 = BenchmarkReturnsQuantileRegimeSpecs(freq='ME')
    else:
        time_period1 = time_period
        regime_params1 = regime_params

    if add_grouped_exposures:
        grouped_exposures_agg, grouped_exposures_by_inst = portfolio_data.get_grouped_long_short_exposures(time_period=time_period1)
        nrows = len(grouped_exposures_agg.keys())
        fig1 = plt.figure(figsize=figsize, constrained_layout=True)
        figs.append(fig1)
        fig1.suptitle(f"{portfolio_data.nav.name} Exposures by groups for period {time_period1.to_str()}",
                     fontweight="bold", fontsize=8, color='blue')
        gs = fig1.add_gridspec(nrows=nrows, ncols=2, wspace=0.0, hspace=0.0)
        local_kwargs = qis.update_kwargs(kwargs=kwargs, new_kwargs=dict(framealpha=0.9))
        for idx, (group, exposures_agg) in enumerate(grouped_exposures_agg.items()):
            datas = {f"{group} aggregated": grouped_exposures_agg[group],
                     f"{group} by instrument": grouped_exposures_by_inst[group]}
            for idx_, (key, df) in enumerate(datas.items()):
                ax = fig1.add_subplot(gs[idx, idx_])
                qis.plot_time_series(df=df,
                                     var_format='{:,.0%}',
                                     legend_stats=qis.LegendStats.AVG_MIN_MAX_LAST,
                                     title=f"{key}",
                                     ax=ax,
                                     **local_kwargs)
                qis.add_bnb_regime_shadows(ax=ax, pivot_prices=time_period1.locate(pivot_prices),
                                           regime_params=regime_params1)
                qis.set_spines(ax=ax, bottom_spine=False, left_spine=False)
                ax.axhline(0, color='black', linewidth=0.5)

    if add_grouped_cum_pnl:
        grouped_pnls_agg, grouped_pnls_by_inst = portfolio_data.get_grouped_cum_pnls(time_period=time_period1)
        nrows = len(grouped_pnls_agg.keys())
        fig1 = plt.figure(figsize=figsize, constrained_layout=True)
        figs.append(fig1)
        fig1.suptitle(f"{portfolio_data.nav.name} P&L by groups for period {time_period1.to_str()}",
                     fontweight="bold", fontsize=8, color='blue')
        gs = fig1.add_gridspec(nrows=nrows, ncols=2, wspace=0.0, hspace=0.0)
        local_kwargs = qis.update_kwargs(kwargs=kwargs, new_kwargs=dict(framealpha=0.9))
        for idx, (group, pnls_agg) in enumerate(grouped_pnls_agg.items()):
            datas = {f"{group} aggregated": grouped_pnls_agg[group],
                     f"{group} by instrument": grouped_pnls_by_inst[group]}
            for idx_, (key, df) in enumerate(datas.items()):
                ax = fig1.add_subplot(gs[idx, idx_])
                qis.plot_time_series(df=df,
                                     var_format='{:,.0%}',
                                     legend_stats=qis.LegendStats.LAST_NONNAN,
                                     title=f"{key}",
                                     ax=ax,
                                     **local_kwargs)
                qis.add_bnb_regime_shadows(ax=ax,
                                           pivot_prices=time_period1.locate(pivot_prices),
                                           regime_params=regime_params1)
                qis.set_spines(ax=ax, bottom_spine=False, left_spine=False)

    return figs



"""
analyse rolling betas of vix to SPY and TLT ETFs
"""
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt
import yfinance as yf
import qis
from qis import PortfolioData

# load VIX ETH
vix = yf.download(tickers=['VXX'], start=None, end=None, ignore_tz=True)['Adj Close'].asfreq('B', method='ffill').rename('Long VIX ETF')

# load becnhmarks benchmarks
benchmark_prices = yf.download(tickers=['SPY', 'TLT'], start=None, end=None, ignore_tz=True)['Adj Close'].asfreq('B', method='ffill')

# create long-only portfolio using vix nav
vix_portfolio = PortfolioData(nav=vix)

# set timeperiod for analysis
time_period = qis.TimePeriod('31Dec2021', None)
perf_params = qis.PerfParams(freq='W-WED', freq_reg='W-WED', alpha_an_factor=52.0,
                             rates_data=yf.download('^IRX', start=None, end=None)['Adj Close'].dropna() / 100.0)
regime_params = qis.BenchmarkReturnsQuantileRegimeSpecs(freq='ME')

prices = pd.concat([vix, benchmark_prices], axis=1).sort_index().dropna()
prices = time_period.locate(prices)

with sns.axes_style("darkgrid"):
    fig, axs = plt.subplots(3, 1, figsize=(10, 10))
    kwargs = dict(framealpha=0.9, fontsize=12, legend_loc='lower left')

    # plot performances
    qis.plot_prices(prices=prices,
                    perf_params=perf_params,
                    title='Log Performance',
                    perf_stats_labels=qis.PerfStatsLabels.DETAILED_WITH_DD.value,
                    is_log=True,
                    ax=axs[0],
                    **kwargs)
    qis.add_bnb_regime_shadows(ax=axs[0], pivot_prices=prices['SPY'], regime_params=regime_params)

    # plot vix betas to benchmarks
    span = 63  # use 3m for half-live
    vix_benchmark_betas = vix_portfolio.compute_portfolio_benchmark_betas(benchmark_prices=benchmark_prices,
                                                                          factor_beta_span=span)
    qis.plot_time_series(df=vix_benchmark_betas,
                         var_format='{:,.2f}',
                         legend_stats=qis.LegendStats.AVG_LAST,
                         title='VIX ETF benchmark multi-variate betas',
                         ax=axs[1],
                         **kwargs)
    qis.add_bnb_regime_shadows(ax=axs[1], pivot_prices=prices['SPY'], regime_params=regime_params)

    # plot performance attribution to betas
    factor_attribution = vix_portfolio.compute_portfolio_benchmark_attribution(benchmark_prices=benchmark_prices,
                                                                               factor_beta_span=span,
                                                                               time_period=time_period)
    qis.plot_time_series(df=factor_attribution.cumsum(axis=0),
                         var_format='{:,.0%}',
                         legend_stats=qis.LegendStats.LAST,
                         title='VIX ETF return attribution to benchmark betas',
                         ax=axs[2],
                         **kwargs)
    pivot_prices = benchmark_prices['SPY'].reindex(index=factor_attribution.index, method='ffill')
    qis.add_bnb_regime_shadows(ax=axs[2], pivot_prices=prices['SPY'], regime_params=regime_params)

    qis.align_x_limits_axs(axs=axs, is_invisible_xs=True)

plt.show()
